# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import json
import os

from tol.core import (
    core_data_object
)
from tol.google_sheets import (
    GoogleSheetDataSource
)


def lr_pacbio():
    gsds = GoogleSheetDataSource({
        'client_secrets': json.loads(os.getenv('GOOGLE_CLIENT_SECRETS')),
        'sheet_key': '1k4_B7_htmlJ762HsoDfZp1lGLJfLuvOX0N9cRh1uQVw',
        'mappings': {
            'sequencing_request': {
                'worksheet_name': 'Lab_Results',
                'columns': {
                    'id': {
                        'heading': 'SANGER SAMPLE ID',
                        'type': 'str'
                    },
                    'library_remaining': {
                        'heading': 'Est. Library\nRemaining\n(ul)',
                        'type': 'str'
                    },
                    'library_remaining_oplc': {
                        'heading': 'Max OPLC From\nRemaining Library\n(@75% recovery)',
                        'type': 'str'
                    },
                    'estimated_max_oplc': {
                        'heading': 'Estimated Maximum OPLC\n(1 cell @50% recovery)',
                        'type': 'int'
                    }
                },
                'header_row': 1,
                'data_start_row': 2
            },
            'sequencing_request_charge': {
                'worksheet_name': 'Charging - Sequencing',
                'columns': {
                    'id': {
                        'heading': 'SANGER SAMPLE ID',
                        'type': 'str'
                    },
                    'portion_of_cell': {
                        'heading': 'Portion of cell',
                        'type': 'str'
                    },
                    'cell_status': {
                        'heading': 'Cell Status',
                        'type': 'str'
                    }
                },
                'header_row': 1,
                'data_start_row': 2
            }

        }
    })
    core_data_object(gsds)
    return gsds



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)


class BenchlingSequencingRequestToElasticSequencingRequestConverter(
        DataObjectToDataObjectOrUpdateConverter):
    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        if data_object.sts_id is not None:
            extraction = None
            tissue_prep = None
            if 'extraction_id' in data_object.attributes:
                extraction = {'id': data_object.extraction_id}
            if 'tissue_prep_id' in data_object.attributes:
                tissue_prep = {'id': data_object.tissue_prep_id}
            ret = self._data_object_factory(
                'sequencing_request',
                data_object.sanger_sample_id,
                attributes={
                    'sample': {'id': str(data_object.sts_id)},
                    'specimen': {'id': str(data_object.specimen_id)},
                    'species': {'id': str(data_object.taxon_id)},
                    'tolid': {'id': data_object.programme_id},
                    'extraction': extraction,
                    'tissue_prep': tissue_prep,
                    **{k: v
                       for k, v in data_object.attributes.items()
                       if k not in ['sanger_sample_id', 'sts_id',
                                    'specimen_id', 'taxon_id', 'extraction_id',
                                    'programme_id', 'tissue_prep_id']}})
            yield ret



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT


def convert_tolqc_samples_to_data_object(samples, target_datasource):
    for sample in samples:
        CoreDataObject = target_datasource.data_object_factory  # noqa N806
        ret = CoreDataObject(
            'sample',
            attributes={
                'sample_ref': sample.name,
                'public_name': sample.specimens.name,
                'common_name': sample.specimens.species.name,
                'supplier_name': sample.specimens.supplied_name,
                'biospecimen_id': sample.specimens.accessions.name,
                'taxon_id': sample.specimens.species.taxon_id
            }
        )
        yield ret



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from abc import ABC, abstractmethod
from collections import defaultdict
from functools import reduce
from itertools import chain
from typing import Any, Dict, Iterator, Optional, Tuple, Type

from sqlalchemy import BinaryExpression, cast, not_
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import MappedColumn, Query

from .model import Model
from .relationship import SqlRelationshipConfig
from ..core import DataSourceFilter


class DatabaseFilter(ABC):
    """Filters an `sqlalchemy.orm` `Query` object"""

    @abstractmethod
    def filter(  # noqa A003
        self,
        query: Query,
        tablename: str,
        model_dict: Dict[str, Type[Model]]
    ) -> Query:
        """Filter the Query object using the given model"""

    @abstractmethod
    def get_column(self, model: Type[Model], key: str) -> MappedColumn:
        """Gets the column for the given `DataObject` key"""

    @abstractmethod
    def apply_joins_on_query(
        self,
        query: Query,
        join_columns: list[MappedColumn]
    ) -> Query:
        """Applys necessary joins on relation columns"""


class DefaultDatabaseFilter(DatabaseFilter):
    """A reasonable-default database filter"""

    # TODO:
    # - sensible error checking/messages (e.g. if column does not exist)

    def __init__(
        self,
        datasource_filter: Optional[DataSourceFilter],
        type_tablename_dict: dict[str, str],
        relationship_config: SqlRelationshipConfig
    ) -> None:

        self.__filter = datasource_filter
        self.__type_tablename_dict = type_tablename_dict
        self.__r_dict = relationship_config.to_dict()
        self.__inverted_dict = {
            v: k for k, v in type_tablename_dict.items()
        }

    def filter(  # noqa A003
        self,
        query: Query,
        tablename: str,
        model_dict: Dict[str, Type[Model]]
    ) -> Query:

        if self.__filter is None:
            return query

        # TODO this is not thread safe
        self.__base_model = model_dict[tablename]
        self.__model_dict = model_dict

        query = self.__join_on_relations(query)

        query = self.__filter_top_and_(query)
        query = self.__filter_top_exact(query)
        query = self.__filter_top_contains(query)
        query = self.__filter_top_in_list(query)
        query = self.__filter_top_range(query)

        return query

    def __join_on_relations(
        self,
        query: Query
    ) -> Query:

        return reduce(
            lambda q, jc: self.apply_joins_on_query(
                q,
                jc
            ),
            self.__get_join_columns_lists(),
            query
        )

    def apply_joins_on_query(
        self,
        query: Query,
        join_columns: list[MappedColumn]
    ) -> Query:

        return reduce(
            lambda q, c: q.join(c),
            join_columns,
            query
        )

    def __get_join_columns_lists(self) -> Iterator[list[MappedColumn]]:
        base_tablename = self.__base_model.get_table_name()

        return (
            list(
                self.__generate_join_columns(
                    base_tablename,
                    k
                )
            )
            for k in self.__generate_relational_keys()
        )

    def __generate_join_columns(
        self,
        base_tablename: str,
        relational_key: str
    ) -> Iterator[MappedColumn]:

        k_split = relational_key.split('.')
        mid_tablename = base_tablename
        mid_type = self.__inverted_dict[mid_tablename]

        for next_ in k_split[:-1]:
            mid_model = self.__model_dict[mid_tablename]
            yield getattr(mid_model, next_)
            mid_type = self.__r_dict[mid_type].to_one[next_]
            mid_tablename = self.__type_tablename_dict[mid_type]

    def __generate_relational_keys(self) -> Iterator[str]:
        chained = chain(
            self.__none_coalesce(self.__filter.exact),
            self.__none_coalesce(self.__filter.contains),
            self.__none_coalesce(self.__filter.in_list),
            self.__none_coalesce(self.__filter.range),
            self.__none_coalesce(self.__filter.and_),
            self.__filter_pointer_targets
        )
        return (
            k for k in chained
            if '.' in k
        )

    @property
    def __filter_pointer_targets(self) -> Iterator[str]:
        not_none = self.__none_coalesce(self.__filter.and_)
        for column_body in not_none.values():
            for term in column_body.values():
                if 'field' in term:
                    yield term['field']

    def __none_coalesce(self, in_: Optional[dict]) -> dict:
        return in_ if in_ is not None else {}

    def __filter_top_and_(self, query: Query) -> Query:
        if not self.__filter.and_:
            return query

        return reduce(
            lambda q, kv: self.__switch_and_term_dict(
                q,
                *kv
            ),
            self.__filter.and_.items(),
            query
        )

    def __switch_and_term_dict(
        self,
        query: Query,
        column_key: str,
        term_dict: dict[str, dict[str, Any]]
    ) -> Query:

        return reduce(
            lambda q, kv: self.__switch_and_term(
                q,
                self.get_column(self.__base_model, column_key),
                *kv
            ),
            term_dict.items(),
            query
        )

    def __switch_and_term(
        self,
        query: Query,
        column: MappedColumn,
        op: str,
        term: dict[str, dict[str, Any]]
    ) -> Query:

        filter_dict = defaultdict(
            lambda: lambda *_: query,
            eq=self.__filter_eq,
            contains=self.__filter_contains,
            in_list=self.__filter_in_list,
            gt=self.__filter_gt,
            gte=self.__filter_gte,
            lt=self.__filter_lt,
            lte=self.__filter_lte,
            exists=self.__filter_exists
        )

        return filter_dict[op](query, column, term)

    def __parse_value_negate(
        self,
        term: dict[str, Any]
    ) -> tuple[Any, bool]:

        negate = term.get('negate', False)

        if 'field' in term:
            field = term['field']
            column = self.get_column(
                self.__base_model,
                field
            )
            return column, negate
        else:
            return term.get('value'), negate

    def __filter_exists(
        self,
        query: Query,
        column: MappedColumn,
        term: dict[str, Any]
    ) -> Query:

        _, negate = self.__parse_value_negate(term)

        if negate:
            return query.filter(
                column.is_(None)
            )
        else:
            return query.filter(
                column.is_not(None)
            )

    def __negatable_filter(
        self,
        query: Query,
        expression: BinaryExpression,
        column: MappedColumn,
        negate: bool
    ) -> Query:

        if negate is True:
            return query.filter(
                (column.is_(None)) | not_(expression)
            )
        else:
            return query.filter(expression)

    def __filter_in_list(
        self,
        query: Query,
        column: MappedColumn,
        term: dict[str, Any]
    ) -> Query:

        value, negate = self.__parse_value_negate(term)
        expression = column.in_(value)

        return self.__negatable_filter(
            query,
            expression,
            column,
            negate
        )

    def __filter_contains(
        self,
        query: Query,
        column: MappedColumn,
        term: dict[str, Any]
    ) -> Query:

        value, negate = self.__parse_value_negate(term)

        if self.__column_is_str(column):
            return self.__filter_contains_str(
                query,
                column,
                value,
                negate
            )
        else:
            return self.__filter_contains_list(
                query,
                column,
                value,
                negate
            )

    def __column_is_str(self, column: MappedColumn) -> bool:
        types = self.__base_model.get_attribute_types()

        return types[column.name] == str

    def __filter_contains_str(
        self,
        query: Query,
        column: MappedColumn,
        value: str,
        negate: bool
    ) -> Query:

        ilike = self.__get_ilike_term(value)
        expression = column.ilike(ilike)

        return self.__negatable_filter(
            query,
            expression,
            column,
            negate
        )

    def __filter_contains_list(
        self,
        query: Query,
        column: MappedColumn,
        value: Any,
        negate: bool
    ) -> Query:

        jsonb_column = cast(column, JSONB)
        expression = jsonb_column.op('@>')([value])

        return self.__negatable_filter(
            query,
            expression,
            column,
            negate
        )

    def __filter_eq(
        self,
        query: Query,
        column: MappedColumn,
        term: dict[str, Any]
    ) -> Query:

        value, negate = self.__parse_value_negate(term)
        expression = column == value

        return self.__negatable_filter(
            query,
            expression,
            column,
            negate
        )

    def __filter_lt(
        self,
        query: Query,
        column: MappedColumn,
        term: dict[str, Any]
    ) -> Query:

        value, negate = self.__parse_value_negate(term)
        expression = column < value

        return self.__negatable_filter(
            query,
            expression,
            column,
            negate
        )

    def __filter_lte(
        self,
        query: Query,
        column: MappedColumn,
        term: dict[str, Any]
    ) -> Query:

        value, negate = self.__parse_value_negate(term)
        expression = column <= value

        return self.__negatable_filter(
            query,
            expression,
            column,
            negate
        )

    def __filter_gt(
        self,
        query: Query,
        column: MappedColumn,
        term: dict[str, Any]
    ) -> Query:

        value, negate = self.__parse_value_negate(term)
        expression = column > value

        return self.__negatable_filter(
            query,
            expression,
            column,
            negate
        )

    def __filter_gte(
        self,
        query: Query,
        column: MappedColumn,
        term: dict[str, Any]
    ) -> Query:

        value, negate = self.__parse_value_negate(term)
        expression = column >= value

        return self.__negatable_filter(
            query,
            expression,
            column,
            negate
        )

    def __filter_top_exact(self, query: Query) -> Query:
        exact_filters = self.__filter.exact
        if exact_filters is None:
            return query
        for k, v in exact_filters.items():
            exact_column = self.get_column(self.__base_model, k)
            query = query.filter(exact_column == v)
        return query

    def __filter_top_contains(self, query: Query) -> Query:
        contains_filters = self.__filter.contains
        if contains_filters is None:
            return query
        for k, v in contains_filters.items():
            contains_column = self.get_column(self.__base_model, k)
            term = self.__get_ilike_term(v)
            query = query.filter(contains_column.ilike(term))
        return query

    def __filter_top_in_list(self, query: Query) -> Query:
        in_filters = self.__filter.in_list
        if in_filters is None:
            return query
        for k, v in in_filters.items():
            in_column = self.get_column(self.__base_model, k)
            query = query.filter(in_column.in_(v))
        return query

    def __filter_top_range(self, query: Query) -> Query:
        range_filters = self.__filter.range
        if range_filters is None:
            return query
        for k, v in range_filters.items():
            range_column = self.get_column(self.__base_model, k)
            from_, to_ = self.__get_between_term(v)
            query = query.filter(range_column.between(from_, to_))
        return query

    def get_column(self, model: Type[Model], key: str) -> MappedColumn:
        if key == 'id':
            return self.__get_id_column(model)
        elif '.' in key:
            return self.__get_relation_column(model, key)
        else:
            return model.get_column(key)

    def __get_id_column(self, model: type[Model]) -> MappedColumn:
        id_key = model.get_id_column_name()
        return model.get_column(id_key)

    def __get_relation_column(
        self,
        model: Type[Model],
        key: str
    ) -> MappedColumn:

        split_keys = key.split('.')

        base_tablename = model.get_table_name()
        mid_type = self.__inverted_dict[base_tablename]

        for split_ in split_keys[:-1]:
            mid_type = self.__r_dict[mid_type].to_one[split_]

        end_tablename = self.__type_tablename_dict[mid_type]
        end_model = self.__model_dict[end_tablename]
        end_key = split_keys[-1]

        return self.get_column(end_model, end_key)

    def __get_ilike_term(self, value: str) -> str:
        escaped = self.__escape_ilike(value)
        return f'%{escaped}%'

    def __get_between_term(self, value: Dict[str, Any]) -> Tuple[Any, Any]:
        from_ = value['from']
        to_ = value['to']
        return from_, to_

    def __escape_ilike(self, value: str) -> str:
        return (
            value.replace('\\', '\\\\').replace('%', '\\%').replace('_', '\\_')
        )



# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import hashlib
import json
import typing
from collections.abc import Callable
from datetime import datetime
from typing import Any, Dict, Iterable, List, Mapping, Optional, Tuple

from cachetools.func import ttl_cache

from caseconverter import (
    kebabcase,
    snakecase
)

from dateutil import parser

from elasticsearch import (Elasticsearch, helpers)

from more_itertools import seekable

from ..core import (
    AttributeMetadata,
    DataId,
    DataObject,
    DataSource,
    DataSourceError,
    DataSourceFilter,
    DefaultAttributeMetadata,
    GroupStatterDataLoader
)
from ..core.operator import (
    Aggregator,
    Counter,
    Cursor,
    DetailGetter,
    Enricher,
    GroupStatter,
    ListGetter,
    PageGetter,
    RelationWriteMode,
    Relational,
    Statter,
    Summariser,
    Updater,
    Upserter,
)
from ..core.operator.updater import DataObjectUpdate
from ..core.relationship import (
    RelationshipConfig
)

if typing.TYPE_CHECKING:
    from ..core.session import OperableSession


class ElasticDataSource(
    DataSource,
    Cursor,
    Summariser,
    DetailGetter,
    Enricher,
    PageGetter,
    ListGetter,
    Aggregator,
    Relational,
    Updater,
    Upserter,
    Counter,
    Statter,
    GroupStatter
):

    def __init__(self, config: Dict,
                 attribute_metadata: AttributeMetadata = DefaultAttributeMetadata):
        super().__init__(config,
                         expected=['uri', 'user', 'password',
                                   'index_prefix', 'relationship_cfg',
                                   'runtime_fields'],
                         attribute_metadata=attribute_metadata)
        """
        relationship_cfg is also supported if we want to handle relationships
        Only FKs pointing to IDs are currently supported
        """
        attribute_metadata.host = self
        self._initialise_elasticsearch()
        self.__lazy = False

    @property
    def lazy_fetch(self) -> bool:
        """
        If `True`, enriched fields will be consulted.

        If `False`, enriched fields will be ignored, and the relation object
        directly fetched by ID, every time.
        """

        return self.__lazy

    @lazy_fetch.setter
    def lazy_fetch(self, new_val: bool) -> None:
        self.__lazy = new_val

    @property
    def _default_write_mode(self) -> RelationWriteMode:
        return RelationWriteMode.FUSED

    def _initialise_elasticsearch(self):
        self.es = Elasticsearch(
            self.uri,
            http_auth=(self.user, self.password),
            timeout=30,
            max_retries=10,
            retry_on_timeout=True
        )
        self.helpers = helpers

    def _convert_data_object_to_dict(self, data_object: DataObject) -> Dict:
        to_ones_dict = {
            k: self._convert_to_one_relation(v)
            for k, v in data_object._to_one_objects.items()
        }
        return data_object.attributes | to_ones_dict

    def _convert_to_one_relation(
        self,
        one_relation: DataObject | None
    ) -> dict[str, Any] | None:

        if one_relation is None:
            return None

        return {
            'id': one_relation.id,
            **one_relation.attributes
        }

    def _prefix_fields(self, dict_: Dict, prefix: str) -> Dict:
        if prefix == '':
            return dict_
        ret = {}
        for k, v in dict_.items():
            ret[prefix + '_' + k] = v
        return ret

    def _add_updated(self, dict_: Dict) -> Dict:
        return {**dict_, 'tol_updated_at': datetime.now().isoformat()}

    def _add_checksum(self, dict_: Dict) -> Dict:
        dhash = hashlib.sha256()
        encoded = json.dumps(dict_, sort_keys=True, default=str).encode()
        dhash.update(encoded)
        return {**dict_, 'checksum': dhash.hexdigest()}

    def _add_uid(self, dict_: Dict, uid: Any) -> Dict:
        return {**dict_, 'uid': f'{uid}'}

    def _convert_dates(self, dict_: Dict) -> Dict:
        ret = {}
        for k, v in dict_.items():
            if isinstance(v, datetime):
                ret[k] = v.isoformat()
            else:
                ret[k] = v
        return ret

    def _stringify_ids(self, dict_: Dict) -> Dict:
        ret = {}
        for k, v in dict_.items():
            if isinstance(v, dict):
                if 'id' in v:
                    v['id'] = str(v['id'])
                ret[k] = self._stringify_ids(v)
            else:
                ret[k] = v

        return ret

    def _action_for_upsert(self, index: str, objects: Iterable[DataObject], id_func: Callable,
                           field_prefix: str):
        for object_ in objects:
            obj = self._convert_data_object_to_dict(object_)
            obj = self._convert_dates(obj)
            obj = self._add_checksum(obj)
            obj = self._add_updated(obj)
            obj = self._prefix_fields(obj, field_prefix)
            obj = self._stringify_ids(obj)
            uid = id_func(object_)
            obj = self._add_uid(obj, uid)
            yield {
                '_op_type': 'update',
                'scripted_upsert': True,
                'upsert': {},
                '_index': index,
                '_id': uid,
                'script': {
                    'source': self._upsert_script,
                    'lang': 'painless',
                    'params': {
                        'upsertWith': obj
                    }
                }
            }

    def get_cursor_page(
        self,
        object_type: str,
        page_size: Optional[int] = None,
        object_filters: Optional[DataSourceFilter] = None,
        search_after: list[str] | None = None,
        session: Optional[OperableSession] = None
    ) -> tuple[Iterable[DataObject], list[str] | None]:

        resp = self.__get_page_response(
            object_type,
            self.update_cursor_filters(
                search_after,
                object_filters
            ),
            'id',
            page_size,
            search_after=search_after
        )

        return self.__format_cursor_response(resp)

    def upsert(
        self,
        object_type: str,
        objects: Iterable[DataObject],
        chunk_size: int = 100,
        id_func=lambda x: x.id,
        field_prefix: str = '',
        **kwargs
    ) -> None:

        index = self.__get_index(object_type)
        (no_of_operations, no_of_errors) = \
            self.helpers.bulk(self.es,
                              self._action_for_upsert(index,
                                                      objects,
                                                      id_func,
                                                      field_prefix),
                              stats_only=True,
                              chunk_size=chunk_size)
        if no_of_errors > 0:
            raise DataSourceError(f'{no_of_errors} errors encountered '
                                  f'upserting {no_of_operations} objects')

    def update(
        self,
        object_type: str,
        updates: Iterable[DataObjectUpdate],
        field_prefix: str = '',
        candidate_key: Iterable[str] = [],
        **kwargs
    ) -> None:

        # This tries to find an object in the DataSource that matches
        # the candidate key. If found it will perform the update

        index = self.__get_index(object_type)
        for (_, update) in updates:
            # We can get the candidate key dynamically from the actual update
            if 'candidate_key_func' in kwargs:
                candidate_key = kwargs['candidate_key_func'](update)
            self.es.update_by_query(
                index=index,
                body=self._action_for_update(index,
                                             update,
                                             field_prefix,
                                             candidate_key),
                wait_for_completion=False
            )

    def _summarise(
        self,
        summary: DataObject,
        ext_and: dict[str, Any] | None = None,
    ) -> None:

        loader = GroupStatterDataLoader(
            self,
            self,
            [],
            summary.source_object_type,
            summary.destination_object_type,
            'Unmanaged summariser (no audit)',
            object_filters=self._mix_in_ext_and(
                summary.object_filters,
                ext_and,
            ),
            group_statter_group_by=summary.group_by,
            group_statter_stats_fields=summary.stats_fields,
            group_statter_stats=summary.stats,
        )
        loader.load(field_prefix=summary.prefix)

    def __format_cursor_response(
        self,
        resp: dict[str, Any]
    ) -> tuple[Iterable[DataObject], list[str] | None]:

        hits = list(resp['hits']['hits'])
        if not hits:
            return [], None

        search_after = hits[-1]['sort']
        objs = self._convert_dict_to_data_objects(hits)

        return objs, search_after

    @property
    def _update_script(self):
        s = """
            for (param in params['upsertWith'].entrySet()) {
                if (param.value != null) {
                    if (ctx._source[param.key] instanceof Map) {
                        for (newParam in param.value.entrySet()) {
                            ctx._source[param.key][newParam.key] = newParam.value;
                        }
                        continue
                    }
                    if (ctx._source[param.key] instanceof ArrayList) {
                        for (newParam in param.value) {
                            if(! ctx._source[param.key].contains(newParam)) {
                                ctx._source[param.key].add(newParam)
                            }
                        }
                        continue
                    }
                }
                ctx._source[param.key] = param.value;
            }
        """
        return s.replace('\n', ' ')

    @property
    def _upsert_script(self):
        s = f"""
            if ( ctx.op == 'create' ) {{
                ctx._source = params['upsertWith']
            }} else {{
                {self._update_script}
            }}
        """
        return s.replace('\n', ' ')

    def _action_for_update(self, index: str, update: Dict,
                           field_prefix: str, candidate_key: Iterable[str]):
        u = self._convert_dates(update)
        u = self._add_checksum(u)
        u = self._add_updated(u)
        f = DataSourceFilter()
        f.and_ = {}
        for key in candidate_key:
            # Don't want key in the upsert as it cannot change anyway
            f.and_[key] = {'eq': {'value': u.pop(key)}}
        u = self._prefix_fields(u, field_prefix)
        query = self._build_elasticsearch_query(
            self.__get_object_type(index),
            object_filters=f)
        return {
            'query': query,
            'script': {
                'source': self._update_script,
                'lang': 'painless',
                'params': {
                    'upsertWith': u
                }
            },
        }

    def __get_index(self, object_type: str) -> str:
        return f'{self.index_prefix}-{kebabcase(object_type)}'

    def __get_object_type(self, index: str) -> str:
        start = len(self.index_prefix) + 1
        return snakecase(index[start:])

    def _field_or_keyword(self, object_type: str, name: str):
        if name == 'id':
            return 'uid.keyword'
        # Runtime fields don't behave the same as text fields
        if object_type in self.runtime_fields and name in self.runtime_fields[object_type]:
            return name
        # An attribute of the object
        if name in self.attribute_types[object_type]:
            field_type = self.attribute_types[object_type][name]
            if field_type == 'str':
                return f'{name}.keyword'
        if '.' in name:
            rc = self.relationship_config[object_type]
            relationship_name, attribute = name.split('.')[0], name.split('.')[1]
            if attribute == 'id':
                return f'{name}.keyword'
            relationship_object_type = rc.to_one[relationship_name]
            attribute_type = self.attribute_types[relationship_object_type][attribute]
            if attribute_type == 'str':
                return f'{name}.keyword'
        return name

    def get_by_id(
        self,
        object_type: str,
        object_ids: Iterable[DataId],
        **kwargs
    ) -> Iterable[DataObject]:
        f = DataSourceFilter()
        f.and_ = {'_id': {'in_list': {'value': object_ids}}}
        # get_by_id is expected to return objects in the order they were asked for
        # or None if not found, hence the following rearrangement.
        seekable_objects = seekable(self.get_list(object_type, object_filters=f))
        for id_ in object_ids:
            seekable_objects.seek(0)
            for obj in seekable_objects:
                if obj.id == id_:
                    yield obj
                    break
            else:
                yield None

    def get_list_page(
        self,
        object_type: str,
        page: int,
        object_filters: DataSourceFilter = None,
        sort_by: str = None,
        page_size: int = None,
        **kwargs
    ) -> Tuple[Iterable[DataObject], int]:

        resp = self.__get_page_response(
            object_type,
            object_filters,
            sort_by,
            page_size,
            page=page
        )

        return (
            self._convert_dict_to_data_objects(
                resp['hits']['hits']
            ),
            resp['hits']['total']['value']
        )

    def __get_page_response(
        self,
        object_type: str,
        object_filters: DataSourceFilter | None,
        sort_by: str | None,
        page_size: int | None,
        page: int | None = None,
        search_after: list[Any] | None = None
    ) -> dict[str, Any]:

        index = self.__get_index(object_type)
        query = self._build_elasticsearch_query(object_type, object_filters)
        sort = self._build_elasticsearch_sort(object_type, sort_by)
        fields = list(self.runtime_fields[object_type].keys()) \
            if object_type in self.runtime_fields else None
        runtime_mappings = self.runtime_fields[object_type] \
            if object_type in self.runtime_fields else None
        if page_size is None:
            page_size = self.get_page_size()
        from_ = (page - 1) * page_size if page is not None else None
        return self.es.search(
            from_=from_,
            size=page_size,
            index=index,
            query=query,
            sort=sort,
            fields=fields,
            runtime_mappings=runtime_mappings,
            search_after=search_after
        )

    def _contains_filter(
        self,
        query: dict[str, Any],
        object_type: str,
        key: str,
        value: str
    ) -> dict[str, Any]:

        search_field = self._field_or_keyword(object_type, key)
        if self.attribute_types[object_type][key] == 'str':
            query['bool']['must'].append(
                {
                    'wildcard': {
                        search_field: {
                            'value': f'{value}*', 'boost': 1.0
                        }
                    }
                }
            )
        else:
            query = self._eq_filter(
                query,
                'must',
                search_field,
                value
            )
        return query

    def _eq_filter(
        self,
        query: dict[str, Any],
        elastic_section: str,
        search_field: str,
        search_value: str
    ) -> dict[str, Any]:

        query['bool'][elastic_section].append({
            'match': {search_field: search_value}
        })

        return query

    def _build_elasticsearch_query(self, object_type: str,
                                   object_filters: DataSourceFilter = None):
        query = {'bool': {'must': [], 'must_not': []}}
        object_filters = self._preprocess_filter(object_type, object_filters)
        # If we want to implement preprocessing of filters, call self._preprocess_filter() here
        if object_filters is None:
            return query
        if object_filters.and_ is not None:
            for k, v in object_filters.and_.items():
                search_field = self._field_or_keyword(object_type, k)
                for op, constraint in v.items():
                    search_value = constraint.get('value')
                    negated = constraint.get('negate', False)
                    elastic_section = 'must_not' if negated else 'must'
                    if 'field' in constraint:
                        other_field = self._field_or_keyword(
                            object_type, constraint['field']
                        )
                        query['bool']['filter'] = \
                            self._get_field_comparison_filter(
                                search_field,
                                other_field,
                                op,
                                negated
                        )
                        continue
                    if op in ['gt', 'gte', 'lt', 'lte']:
                        query['bool'][elastic_section].append({
                            'range': {search_field: {op: search_value}}
                        })
                    if op in ['eq']:
                        query = self._eq_filter(
                            query,
                            elastic_section,
                            search_field,
                            search_value
                        )
                    if op in ['contains']:
                        query['bool'][elastic_section].append({
                            'wildcard': {
                                search_field: {'value': f'{search_value}*', 'boost': 1.0}
                            }
                        })
                    if op in ['exists']:
                        query['bool'][elastic_section].append({
                            'exists': {'field': search_field}
                        })
                    if op in ['in_list']:
                        query['bool'][elastic_section].append({
                            'terms': {search_field: search_value, 'boost': 1.0}
                        })
        return query

    def _get_field_comparison_filter(self, field1: str, field2: str, op: str, negated: bool) -> \
            Dict[str, Dict[str, str]]:
        op_mappings = {
            'eq': '==',
            'lt': '<',
            'lte': '<=',
            'gt': '>',
            'gte': '>='
        }
        negated_mappings = {  # What to return if negated
            True: 'true',
            False: 'false'
        }
        # return {negated_mappings[not negated]}
        return {
            'script': {
                'script': {
                    'source': f"""
                        if (doc[params['field1']].size() > 0
                            && doc[params['field2']].size() > 0) {{
                            if (doc[params['field1']].value.compareTo(doc[params['field2']].value)
                                {op_mappings[op]} 0) {{
                                return {negated_mappings[not negated]}
                            }}
                        }}
                        return {negated_mappings[negated]};
                    """,
                    'params': {
                        'field1': field1,
                        'field2': field2
                    }
                }
            }
        }

    def _build_elasticsearch_sort(
        self,
        object_type: str,
        sort_by: str
    ) -> list[dict[str, str]]:
        default_sort = {'uid.keyword': 'asc'}
        if sort_by is None:
            return [default_sort]
        if sort_by == '-id':
            return self.__build_uid_sort(True)
        if sort_by == 'id':
            return self.__build_uid_sort(False)
        return [
            self.__build_sort(object_type, sort_by),
            default_sort
        ]

    def __build_uid_sort(
        self,
        desc: bool
    ) -> list[dict[str, str]]:

        order = 'desc' if desc else 'asc'
        return [
            {
                'uid.keyword': {
                    'order': order,
                    'unmapped_type': 'keyword'
                }
            }
        ]

    def __build_sort(
        self,
        object_type: str,
        sort_by: str
    ) -> dict[str, str]:

        if sort_by.startswith('-'):
            field = self._field_or_keyword(object_type, sort_by[1:])
            order = 'desc'
        else:
            field = self._field_or_keyword(object_type, sort_by)
            order = 'asc'

        return {field: order}

    def get_list(
        self,
        object_type: str,
        object_filters: DataSourceFilter | None = None,
        session: OperableSession | None = None,
        **kwargs
    ) -> Iterable[DataObject]:
        index = self.__get_index(object_type)
        query = self._build_elasticsearch_query(object_type, object_filters)
        fields = list(self.runtime_fields[object_type].keys()) \
            if object_type in self.runtime_fields else None
        runtime_mappings = self.runtime_fields[object_type] \
            if object_type in self.runtime_fields else None
        generator = self.helpers.scan(self.es,
                                      index=index,
                                      scroll='10m',
                                      size=500,
                                      query={'query': query},
                                      fields=fields,
                                      runtime_mappings=runtime_mappings)
        return self._convert_dict_to_data_objects(generator)

    def _convert_dict_to_data_objects(self, objs: Dict) -> Iterable:
        for obj in objs:
            if '_source' in obj:
                type_ = self.__get_object_type(obj['_index'])
                id_ = obj['_id']
                attributes = obj['_source']
                runtime_attributes = obj['fields'] if 'fields' in obj else {}
                yield self._convert_data_dict_to_data_object(
                    type_,
                    id_,
                    attributes,
                    runtime_attributes
                )
            else:
                yield None

    def _convert_data_dict_to_data_object(self, type_, id_, data, runtime_data):
        attributes = {
            k: self.__make_dates(type_, k, v) for k, v in data.items()
            if k in self.attribute_types[type_]
        }
        runtime_attributes = {
            k: self.__make_dates(type_, k, v[0]) for k, v in runtime_data.items()
            if k in self.attribute_types[type_]
        }
        to_one = self.__make_to_one_relations(type_, data)
        return self.data_object_factory(
            type_,
            id_=id_,
            attributes=attributes | runtime_attributes,
            to_one=to_one
        )

    def __make_to_one_relations(
        self,
        type_: str,
        data: dict[str, Any]
    ) -> dict[str, Optional[DataObject]]:

        if type_ not in self.relationship_config:
            return {}

        if self.relationship_config[type_].to_one is None:
            return {}

        return {
            k: self.__make_to_one_relation(data.get(k), v)
            for k, v in self.relationship_config[type_].to_one.items()
        }

    def __make_to_one_relation(
        self,
        relation_data: Optional[dict[str, Any]],
        type_: str
    ) -> Optional[DataObject]:

        if (
            relation_data is None
            or not isinstance(relation_data, Mapping)
        ):
            return None

        id_ = relation_data.get('id')

        if id_ is None:
            return None

        return self._convert_data_dict_to_data_object(
            type_,
            id_,
            relation_data,
            {}  # This can be empty because runtime_fields are not applicable for enriched objects
        )

    def __make_dates(self, object_type, attribute_name, value):
        if self.attribute_types[object_type][attribute_name] == 'datetime' and \
                isinstance(value, str):
            return parser.parse(value)
        return value

    def get_aggregations(
        self,
        object_type: str,
        aggregations: Dict,
        object_filters: DataSourceFilter = None,
        **kwargs
    ) -> Dict:
        index = self.__get_index(object_type)
        query = self._build_elasticsearch_query(object_type, object_filters)
        fields = list(self.runtime_fields[object_type].keys()) \
            if object_type in self.runtime_fields else None
        runtime_mappings = self.runtime_fields[object_type] \
            if object_type in self.runtime_fields else None
        resp = self.es.search(
            size=0,
            index=index,
            query=query,
            aggregations=aggregations,
            fields=fields,
            runtime_mappings=runtime_mappings
        )
        return resp['aggregations']

    def get_stats(
        self,
        object_type: str,
        stats_fields: List[str] = [],
        stats: List[str] = [],
        object_filters: DataSourceFilter = None,
        **kwargs
    ):
        aggs = self.__get_aggs(
            object_type=object_type,
            stats_fields=stats_fields,
            stats=stats)
        agg_results = self.get_aggregations(
            object_type=object_type,
            aggregations=aggs,
            object_filters=object_filters
        )
        return self.__get_data_from_stats_aggregation(
            aggregation_result=agg_results,
            object_type=object_type,
            stats_fields=stats_fields,
            stats=stats
        )

    def __get_data_from_stats_aggregation(
            self,
            aggregation_result,
            object_type,
            stats_fields,
            stats
    ):
        stats_values = {}
        for stats_field in stats_fields:
            stats_values[stats_field] = {}
            for stat in stats:
                stat_value = aggregation_result[f'{stats_field}_{stat}']['value']
                python_type = self.attribute_types[object_type][stats_field]
                if python_type == 'datetime' and stat_value is not None \
                        and stat in ['min', 'max']:
                    stat_value = datetime.fromtimestamp(stat_value / 1000)
                stats_values[stats_field][stat] = stat_value
        return {'stats': stats_values}

    def get_group_stats(
        self,
        object_type: str,
        group_by: List[str],
        stats_fields: List[str] = [],
        stats: List[str] = [],
        object_filters: DataSourceFilter = None,
        **kwargs
    ) -> Iterable[dict[Any, int]]:

        after_key = None
        while True:
            after_key, buckets = self.__get_group_stats_page(
                object_type,
                group_by,
                stats_fields=stats_fields,
                stats=stats,
                object_filters=object_filters,
                after_key=after_key)
            if len(buckets) == 0:
                break
            yield from buckets

    def __get_group_stats_page(
        self,
        object_type: str,
        group_by: List[str],
        stats_fields: List[str] = [],
        stats: List[str] = [],
        after_key: str = None,
        object_filters: DataSourceFilter = None,
    ):
        # This will return a potentially large set of results, so we need
        # to page through them
        aggregation = {
            'counts': {
                'composite': {
                    'sources': [{
                        field: {
                            'terms': {
                                'field': self._field_or_keyword(object_type, field)
                            }
                        }
                    } for field in group_by]
                },
            }
        }
        if stats_fields is not None:
            aggregation['counts']['aggregations'] = self.__get_aggs(
                object_type,
                stats_fields,
                stats
            )
        if after_key is not None:
            aggregation['counts']['composite']['after'] = after_key
        agg_page = self.get_aggregations(
            object_type,
            aggregations=aggregation,
            object_filters=object_filters)
        after_key, buckets = self.__get_data_from_group_stats_aggregation(
            agg_page,
            object_type,
            stats_fields,
            stats
        )
        return after_key, buckets

    def __get_aggs(
            self,
            object_type: str,
            stats_fields: List,
            stats: List
    ):
        ret = {}
        for stats_field in stats_fields:
            for stat in stats:
                agg = {stat: {'field': self._field_or_keyword(object_type, stats_field)}}
                if stat == 'union':
                    # This is a bespoke aggregation
                    agg = self.__get_union_aggregation(object_type, stats_field)
                elif stat == 'unique':
                    agg = self.__get_unique_count_aggregation(object_type, stats_field)
                elif self.attribute_types[object_type][stats_field] == 'str' \
                        and stat in ['min', 'max']:
                    agg = self.__get_string_aggregation(object_type, stats_field, stat)
                ret[f'{stats_field}_{stat}'] = agg
        return ret

    def __get_data_from_group_stats_aggregation(
            self,
            aggregation_result,
            object_type,
            stats_fields,
            stats
    ):
        # The after_key is sent back in one request and we use it as-is in the next request
        after_key = None
        if 'after_key' in aggregation_result['counts']:
            after_key = aggregation_result['counts']['after_key']
        buckets = aggregation_result['counts']['buckets']
        # all_stats looks like:
        # [{'key': {'first_group_by': 'value_of_first_group_by'}
        #           'second_group_by': 'value_of_second_group_by'},
        #   'stats': {'count': 123,
        #             'stats_field_stat': 345}}]
        all_stats = []
        for v in buckets:
            stats_values = {'count': v['doc_count']}
            for stats_field in stats_fields:
                stats_values[stats_field] = {}
                for stat in stats:
                    stat_value = v[f'{stats_field}_{stat}']['value']
                    python_type = self.attribute_types[object_type][stats_field]
                    if python_type == 'datetime' and stat_value is not None \
                            and stat in ['min', 'max']:
                        stat_value = datetime.fromtimestamp(stat_value / 1000)
                    stats_values[stats_field][stat] = stat_value
            all_stats.append({'key': v['key'], 'stats': stats_values})

        return after_key, all_stats

    def __get_union_aggregation(self, object_type, field):
        """
        This function is building up a union of all elements of a list in
        the aggregated field
        init_script: This is what is run at the start of each bucket
        map_script: This builds up a list, PER SHARD, of the elements in all
        records in the bucket
        combine_script: This just returns the per-shard list in our case
        reduce_script: This combines the per-shard lists into the final list
        See information on scripted metrics for Elastic for more details
        """
        field_or_keyword = self._field_or_keyword(object_type, field)
        agg = {
            'scripted_metric': {
                'init_script': 'state.list = []',
                'map_script': f"""
                    for (element in doc['{field_or_keyword}']) {{
                        if (!state.list.contains(element)) {{
                            state.list.add(element)
                        }}
                    }}
                """,
                'combine_script': 'return state.list',
                'reduce_script': """
                    ArrayList ret = [];
                    for (a in states) {
                        for (element in a) {
                            if (!ret.contains(element)) {
                                ret.add(element)
                            }
                        }
                    }
                    return ret;
                """
            }
        }
        return agg

    def __get_string_aggregation(self, object_type, field, stat):
        """
        This function is calculating the min and max of a string

        """
        field_or_keyword = self._field_or_keyword(object_type, field)
        comparator = '>'
        if stat == 'min':
            comparator = '<'
        agg = {
            'scripted_metric': {
                'init_script': 'state.stat = null',
                'map_script': f"""
                    for (ss in doc['{field_or_keyword}']) {{
                        if (state.stat == null) {{
                            state.stat = ss; continue
                        }}
                        if (ss.compareTo(state.stat) {comparator} 0) {{
                            state.stat = ss
                        }}
                    }}
                    """,
                'combine_script': 'return state.stat',
                'reduce_script': f"""
                    String ret = null;
                    for (a in states) {{
                        if (a == null) {{
                            continue
                        }}
                        if (ret == null) {{
                            ret = a;
                            continue
                        }}
                        if (a.compareTo(ret) {comparator} 0) {{
                            ret = a
                        }}
                    }}
                    return ret;
                """
            }
        }
        return agg

    def __get_unique_count_aggregation(self, object_type, field):
        """
        This function is calculating the unique values in the given field

        """
        field_or_keyword = self._field_or_keyword(object_type, field)
        agg = {
            'scripted_metric': {
                'params': {
                    'fieldName': field_or_keyword
                },
                'init_script': 'state.list = []',
                'map_script': """
                    if(doc[params.fieldName].size() > 0) {
                        state.list.add(doc[params.fieldName].value);
                    }
                    """,
                'combine_script': 'return state.list;',
                'reduce_script': """
                    Map uniqueValueMap = new HashMap();
                    int count = 0;
                    for(shardList in states) {
                        if(shardList != null) {
                            for(key in shardList) {
                                if(!uniqueValueMap.containsKey(key)) {
                                    count +=1;
                                    uniqueValueMap.put(key, key);
                                }
                            }
                        }
                    }
                    return count;
                """
            }
        }
        return agg

    def get_count(
        self,
        object_type: str,
        object_filters: DataSourceFilter = None,
        **kwargs
    ) -> int:
        index = self.__get_index(object_type)
        query = self._build_elasticsearch_query(object_type, object_filters)
        fields = list(self.runtime_fields[object_type].keys()) \
            if object_type in self.runtime_fields else None
        runtime_mappings = self.runtime_fields[object_type] \
            if object_type in self.runtime_fields else None
        # We are not using es.count so that we can use runtime fields
        resp = self.es.search(
            index=index,
            track_total_hits=True,
            size=0,
            query=query,
            fields=fields,
            runtime_mappings=runtime_mappings
        )
        return resp['hits']['total']['value']

    @property
    @ttl_cache(ttl=3600)
    def supported_types(self):
        index_names = self.es.cat.indices(h='index', s='index').split()
        return [self.__get_object_type(index_name)
                for index_name in index_names
                if index_name.startswith(self.index_prefix)]

    def __map_type(self, type_: str) -> str:
        if type_ in ['text', 'keyword']:
            return 'str'
        if type_ == 'long':
            return 'int'
        if type_ == 'date':
            return 'datetime'
        return type_

    def _get_attribute_types_for_object_type(self, object_type: str) -> Dict:
        index_name = self.__get_index(object_type)
        mapping = self.es.indices.get_mapping(index=index_name)
        if 'properties' not in mapping[index_name]['mappings']:
            return {}
        properties = mapping[index_name]['mappings']['properties']
        standard_types = {
            property_name: self.__map_type(properties[property_name]['type'])
            for property_name in properties
            if 'type' in properties[property_name]
            and 'type' != 'uid'
        }
        runtime_types = {
            name: self.__map_type(self.runtime_fields[object_type][name]['type'])
            for name in self.runtime_fields[object_type].keys()
        } if object_type in self.runtime_fields else {}
        return standard_types | runtime_types

    @property
    @ttl_cache(ttl=3600)
    def attribute_types(self) -> dict[str, dict[str, str]]:
        return {
            t: self._get_attribute_types_for_object_type(t)
            for t in self.supported_types
        }

    @property
    def relationship_config(self) -> dict[str, RelationshipConfig]:
        return self.relationship_cfg

    def get_to_one_relation(
        self,
        source: DataObject,
        relationship_name: str,
        **kwargs
    ) -> Optional[DataObject]:

        self.__validate_to_one_relation(source)

        if self.lazy_fetch and relationship_name in source._to_one_objects:
            return source._to_one_objects.get(relationship_name)

        to_one = self.relationship_config[source.type].to_one

        if relationship_name not in to_one:
            raise DataSourceError('Bad relationship name')

        new_source: DataObject = self.get_one(source.type, source.id)

        local_relation = new_source._to_one_objects.get(relationship_name)
        if local_relation is None:
            return None

        if self.lazy_fetch:
            return local_relation

        target_type = to_one[relationship_name]
        target_id = local_relation.id
        return self.get_one(target_type, target_id)

    def get_to_many_relations(
        self,
        source: DataObject,
        relationship_name: str,
        **kwargs
    ) -> Iterable[DataObject]:
        if self.relationship_config is None:
            raise DataSourceError('There are no relationships defined')
        relationship_config = self.relationship_config[source.type]
        related_object_type = relationship_config.to_many[relationship_name]
        related_object_fk_attribute = relationship_config.foreign_keys[relationship_name]

        # Get all the related objects that point to this source object
        f = DataSourceFilter()
        f.and_ = {related_object_fk_attribute: {'eq': {'value': source.id}}}
        related_objects = self.get_list(related_object_type, object_filters=f)
        return related_objects

    def __validate_to_one_relation(self, source: DataObject) -> None:
        if self.relationship_config is None:
            raise DataSourceError('There are no relationships defined')

        if source.type not in self.relationship_config:
            raise DataSourceError('This type has no relationships')



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)
from ...core.operator.updater import DataObjectUpdate


class ElasticTolidToElasticGenomeNoteUpdateConverter(
        DataObjectToDataObjectOrUpdateConverter):

    def convert(self, data_object: DataObject) -> Iterable[DataObjectUpdate]:
        species = data_object.to_one_relationships['tolid_species']
        yield (
            None,
            {
                'gn_species': {'id': species.id},
                'gn_tolid.id': data_object.id
            }
        )



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from abc import ABC, abstractmethod
from typing import Any, Iterable

from .detail_getter import DetailGetter
from .relational import Relational
from ..datasource_filter import DataSourceFilter


if typing.TYPE_CHECKING:
    from ..data_object import DataObject


class Summariser(
    DetailGetter,
    Relational,
    ABC
):

    @abstractmethod
    def _summarise(
        self,
        summary_object: DataObject,
        ext_and: dict[str, Any] | None = None,
    ) -> None:
        """
        Summarises according to the given `DataObject`
        summary instance.
        """

    def summarise_all(
        self,
        summary_objects: Iterable[DataObject],
    ) -> None:
        """
        Summarises, across all types, using the given summary
        object-config instances.
        """

        for obj in summary_objects:
            self._summarise(obj)

    def summarise_type(
        self,
        summary_objects: Iterable[DataObject],
        source_object_type: str,
    ) -> None:
        """
        Summarises, for only the given `object_type`, using the given summary
        object-config instances.
        """

        filtered_summaries = self._filter_by_source_type(
            summary_objects,
            source_object_type,
        )

        for obj in filtered_summaries:
            self._summarise(obj)

    def resummarise_by_ids(
        self,
        summary_objects: Iterable[DataObject],
        source_object_type: str,
        source_object_ids: Iterable[str],
    ) -> None:
        """
        More restrictive than `summarise_all()`

        Re-summarises, using the given summary instances, only the set of changes
        affecting the `DataObject` instances of given `object_type` and
        `object_ids`.

        Returns None, but performs summarization operations.
        """
        objects_to_summarise, _ = self.get_objects_to_summarise(
            summary_objects,
            source_object_type,
            source_object_ids,
        )

        for summary_obj, ext_and in objects_to_summarise:
            self._summarise(
                summary_obj,
                ext_and=ext_and,
            )

    def _mix_in_ext_and(
        self,
        object_filters: dict[str, Any] | None,
        ext_and: dict[str, Any] | None,
    ) -> DataSourceFilter:

        if not ext_and:
            return DataSourceFilter(
                and_=object_filters,
            )

        if not object_filters:
            return DataSourceFilter(
                and_=ext_and,
            )

        return DataSourceFilter(
            object_filters | ext_and,
        )

    def _filter_by_source_type(
        self,
        summary_objects: Iterable[DataObject],
        source_object_type: str,
    ) -> list[DataObject]:

        return [
            s for s in summary_objects
            if s.source_object_type == source_object_type
        ]

    def get_objects_to_summarise(
        self,
        summary_objects: Iterable[DataObject],
        source_object_type: str,
        source_object_ids: Iterable[str],
    ) -> tuple[list[tuple[DataObject, dict[str, Any]]], dict[str, list[str]]]:
        """
        Helper function for resummariser_by_ids that returns the objects to be summarised
        along with their extended filter conditions.

        Returns:
            A tuple containing:
            - A list of tuples containing (summary_object, ext_and) pairs that can be used
            for summarisation.
            - A list of relationship IDs that were extracted during the process, which can
            be used for subsequent operations like enrichment.
        """
        filtered_summaries = self._filter_by_source_type(
            summary_objects,
            source_object_type,
        )

        if not filtered_summaries:
            return [], {}

        source_objs = list(
            self.get_by_ids(
                source_object_type,
                source_object_ids,
            )
        )

        result = []
        relationship_ids_by_type = {}

        for s in filtered_summaries:
            first_group_by: str = s.group_by[0]
            relationship_hops = '.'.join(
                first_group_by.split('.')[:-1]
            )
            relationship_id_target = f'{relationship_hops}.id'

            relationship_ids_raw = (
                o.get_field_by_name(relationship_id_target)
                for o in source_objs
                if o is not None
            )
            relationship_ids: list[str] = [
                i for i in relationship_ids_raw
                if i is not None
            ]

            dest_type = s.destination_object_type
            if dest_type not in relationship_ids_by_type:
                relationship_ids_by_type[dest_type] = []
            relationship_ids_by_type[dest_type].extend(relationship_ids)

            ext_and = {
                relationship_id_target: {
                    'in_list': {
                        'value': relationship_ids
                    }
                }
            }

            result.append((s, ext_and))

        for obj_type in relationship_ids_by_type:
            relationship_ids_by_type[obj_type] = list(set(relationship_ids_by_type[obj_type]))

        return result, relationship_ids_by_type



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT


def convert_tolid_species_to_data_object(species, target_datasource):
    CoreDataObject = target_datasource.data_object_factory  # noqa N806
    for s in species:
        ret = CoreDataObject('species', {
            'taxonomy_id': s.taxonomy_id,
            'prefix': s.prefix,
            'name': s.name,
            'common_name': s.common_name,
            'genus': s.genus,
            'family': s.family,
            'tax_order': s.tax_order,
            'tax_class': s.tax_class,
            'phylum': s.phylum,
            'kingdom': s.kingdom
        })
        yield ret



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)


class LrpacbioSequencingRequestToElasticSequencingRequestConverter(
        DataObjectToDataObjectOrUpdateConverter):

    def _clean_attribute(self, value):
        if isinstance(value, str):
            try:
                return max(float(value), 0.0)
            except ValueError:
                return None
        else:
            return max(value, 0.0)

    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        target_attributes = {}

        for field in ['library_remaining', 'library_remaining_oplc', 'estimated_max_oplc']:
            if (field in data_object.attributes
                    and data_object.attributes[field] is not None):
                target_attributes[field] = self._clean_attribute(
                    data_object.attributes[field]
                )

        ret = self._data_object_factory(
            'sequencing_request',
            data_object.id,
            attributes=target_attributes
        )
        return iter([ret])



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT


def convert_benchling_samples_to_data_object(samples, target_datasource):
    CoreDataObject = target_datasource.data_object_factory  # noqa N806
    for s in samples:
        if s.sts_id is not None:
            yield CoreDataObject(s)



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from .convert_bioscan_barcoding_run_data_to_data_object import convert_bioscan_barcoding_run_data_to_data_object  # noqa
from .convert_bold_samples_to_data_object import convert_bold_samples_to_data_object  # noqa
from .convert_to_data_objects import convert_to_data_objects  # noqa
from .convert_tolid_species_to_data_object import convert_tolid_species_to_data_object  # noqa
from .convert_tolid_specimen_to_data_object import convert_tolid_specimen_to_data_object  # noqa
from .convert_tolqc_run_data_to_data_object import convert_tolqc_run_data_to_data_object  # noqa
from .convert_tolqc_samples_to_data_object import convert_tolqc_samples_to_data_object  # noqa
from .convert_treeofsex_species_to_data_object import convert_treeofsex_species_to_data_object  # noqa
from .convert_specimen_status_to_data_object import convert_specimen_status_to_data_object  # noqa
from .convert_species_lab_work_status_to_data_object import convert_species_lab_work_status_to_data_object  # noqa
from .convert_portaldb_sequencing_request_events_to_sequencing_requests import convert_portaldb_sequencing_request_events_to_sequencing_requests  # noqa
from .convert_benchling_samples_to_data_object import convert_benchling_samples_to_data_object  # noqa
from .convert_elastic_samples_to_barcoding_run_data import convert_elastic_samples_to_barcoding_run_data  # noqa



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import os

from .portal_attributes import portal_attributes
from ..core import (
    core_data_object
)
from ..core.data_source_attribute_metadata import data_source_attribute_metadata
from ..core.relationship import RelationshipConfig
from ..elastic import (
    ElasticDataSource,
    RuntimeField,
    RuntimeFields
)


def elastic():
    rc_run_data = RelationshipConfig()
    rc_run_data.to_one = {'benchling_extraction': 'extraction',
                          'benchling_sample': 'sample',
                          'mlwh_sequencing_request': 'sequencing_request',
                          'mlwh_specimen': 'specimen',
                          'mlwh_species': 'species',
                          'mlwh_tolid': 'tolid',
                          'tolqc_sequencing_request': 'sequencing_request',
                          'tolqc_specimen': 'specimen',
                          'tolqc_species': 'species',
                          'tolqc_tolid': 'tolid'}

    rc_sequencing_request = RelationshipConfig()
    rc_sequencing_request.to_one = {'benchling_extraction': 'extraction',
                                    'benchling_sample': 'sample',
                                    'benchling_species': 'species',
                                    'benchling_tolid': 'tolid',
                                    'benchling_specimen': 'specimen',
                                    'benchling_tissue_prep': 'tissue_prep',
                                    'mlwh_species': 'species',
                                    'mlwh_tolid': 'tolid',
                                    'mlwh_specimen': 'specimen',
                                    'mlwh_sample': 'sample'}
    rc_sequencing_request.to_many = {
        'mlwh_run_datas': 'run_data',
        'tolqc_run_datas': 'run_data'
    }
    rc_sequencing_request.foreign_keys = {
        'mlwh_run_datas': 'mlwh_sequencing_request.id',
        'tolqc_run_datas': 'tolqc_sequencing_request.id'
    }

    rc_extraction = RelationshipConfig()
    rc_extraction.to_one = {'benchling_sample': 'sample',
                            'benchling_species': 'species',
                            'benchling_specimen': 'specimen',
                            'benchling_tolid': 'tolid',
                            'benchling_tissue_prep': 'tissue_prep'}
    rc_extraction.to_many = {
        'benchling_sequencing_requests': 'sequencing_request'
    }
    rc_extraction.foreign_keys = {
        'benchling_sequencing_request': 'benchling_extraction.id'
    }

    rc_sample = RelationshipConfig()
    rc_sample.to_one = {
        'sts_specimen': 'specimen',
        'benchling_specimen': 'specimen',
        'sts_species': 'species',
        'benchling_species': 'species',
        'sts_tolid': 'tolid',
        'tolid_tolid': 'tolid',
        'benchling_tolid': 'tolid',
        'sts_manifest': 'manifest',
        'sts_sampleset': 'sampleset'
    }
    rc_sample.to_many = {
        'benchling_sequencing_requests': 'sequencing_request',
        'benchling_tissue_preps': 'tissue_prep'
    }
    rc_sample.foreign_keys = {
        'benchling_sequencing_requests': 'benchling_sample.id',
        'benchling_tissue_preps': 'benchling_sample.id'
    }

    rc_sampleset = RelationshipConfig()
    rc_sampleset.to_one = {}
    rc_sampleset.to_many = {
        'sts_manifests': 'manifest',
        'sts_samples': 'sample'
    }
    rc_sampleset.foreign_keys = {
        'sts_manifests': 'sts_sampleset.id',
        'sts_samples': 'sts_sampleset.id'
    }

    rc_manifest = RelationshipConfig()
    rc_manifest.to_one = {
        'sts_sampleset': 'sampleset'
    }
    rc_manifest.to_many = {'sts_samples': 'sample'}
    rc_manifest.foreign_keys = {
        'sts_samples': 'sts_manifest.id'
    }

    rc_tolid = RelationshipConfig()
    rc_tolid.to_one = {'informatics_specimen': 'specimen',
                       'tolid_specimen': 'specimen',
                       'tolid_species': 'species'}
    rc_tolid.to_many = {
        'benchling_tissue_preps': 'tissue-prep',
        'grit_curations': 'curation',
        'gap_assemblies': 'assembly',
        'gn_genome_notes': 'genome_note'
    }
    rc_tolid.foreign_keys = {
        'benchling_tissue_preps': 'benchling_tolid.id',
        'grit_curations': 'grit_tolid.id',
        'gap_assemblies': 'gap_tolid.id',
        'gn_genome_notes': 'gn_tolid.id'
    }

    rc_specimen = RelationshipConfig()
    rc_specimen.to_many = {
        'benchling_extractions': 'extraction',
        'benchling_samples': 'sample',
        'benchling_sequencing_request': 'sequencing_request',
        'mlwh_sequencing_request': 'sequencing_request',
        'sts_samples': 'sample',
    }
    rc_specimen.foreign_keys = {
        'benchling_extractions': 'benchling_specimen.id',
        'benchling_samples': 'benchling_specimen.id',
        'benchling_sequencing_request': 'benchling_specimen.id',
        'mlwh_sequencing_request': 'mlwh_specimen.id',
        'sts_samples': 'sts_specimen.id',
    }

    rc_species = RelationshipConfig()
    rc_species.to_many = {'sts_samples': 'sample',
                          'benchling_samples': 'sample',
                          'benchling_tissue_preps': 'tissue_prep',
                          'grit_curations': 'curation',
                          'gap_assemblies': 'assembly',
                          'gn_genome_notes': 'genome_note'}
    rc_species.foreign_keys = {
        'sts_samples': 'sts_species.id',
        'benchling_samples': 'benchling_species.id',
        'benchling_tissue_preps': 'benchling_species.id',
        'grit_curations': 'grit_species.id',
        'gap_assemblies': 'gap_species.id',
        'gn_genome_notes': 'gn_species.id'
    }

    rc_tissue_prep = RelationshipConfig()
    rc_tissue_prep.to_one = {'benchling_species': 'species',
                             'benchling_sample': 'sample',
                             'benchling_specimen': 'specimen',
                             'benchling_tolid': 'tolid'}
    rc_tissue_prep.to_many = {
        'benchling_extractions': 'extraction',
        'benchling_sequencing_requests': 'sequencing_request',
        'benchling_tissue_preps': 'tissue_prep'
    }
    rc_tissue_prep.foreign_keys = {
        'benchling_extractions': 'benchling_tissue_prep.id',
        'benchling_sequencing_requests': 'benchling_tissue_prep.id',
        'benchling_tissue_preps': 'benchling_tissue_prep.id'
    }

    rc_curation = RelationshipConfig()
    rc_curation.to_one = {'grit_species': 'species',
                          'grit_tolid': 'tolid'}

    rc_assembly = RelationshipConfig()
    rc_assembly.to_one = {'gap_species': 'species'}
    rc_assembly.to_many = {
        'gn_genome_notes': 'genome_note',
        'gap_assembly_analyses': 'assembly_analysis'
    }
    rc_assembly.foreign_keys = {
        'gn_genome_notes': 'gn_assembly.id',
        'gap_assembly_analyses': 'gap_assembly.id'
    }

    rc_assembly_analysis = RelationshipConfig()
    rc_assembly_analysis.to_one = {
        'gap_species': 'species',
        'gap_assembly': 'assembly'
    }

    rc_genome_note = RelationshipConfig()
    rc_genome_note.to_one = {
        'gn_assembly': 'assembly',
        'gn_species': 'species',
        'gn_tolid': 'tolid'
    }

    relationship_config = {'run_data': rc_run_data,
                           'sequencing_request': rc_sequencing_request,
                           'extraction': rc_extraction,
                           'sample': rc_sample,
                           'sampleset': rc_sampleset,
                           'manifest': rc_manifest,
                           'tolid': rc_tolid,
                           'specimen': rc_specimen,
                           'species': rc_species,
                           'tissue_prep': rc_tissue_prep,
                           'curation': rc_curation,
                           'assembly': rc_assembly,
                           'assembly_analysis': rc_assembly_analysis,
                           'genome_note': rc_genome_note}

    runtime_fields = {
        'species': {
            'calc_done_date': RuntimeFields.latest_date(
                [
                    'mlwh_run_data_mlwh_run_complete_rnaseq_min',
                    'grit_curation_grit_in_submission_date_min'
                ],
                allow_missing=False
            ),
            'calc_is_novel': RuntimeField(
                field_type='boolean',
                dependencies=['sts_species_id'],
                function_body="""
                    emit(!(doc.containsKey('sts_sample_sts_tollab_assign_date_min')
                        && doc['sts_sample_sts_tollab_assign_date_min'].size() > 0))
                """,
                function_default='emit(false)'
            ).to_dict(),
            'calc_pm_status': {
                'type': 'keyword',
                'script': {
                    'source': """
                        String status = null;
                        if (doc.containsKey
                            ('informatics_tolid_informatics_status_summary_min.keyword')
                            && doc['informatics_tolid_informatics_status_summary_min.keyword']
                            .size() > 0) {
                            status =
                            doc['informatics_tolid_informatics_status_summary_min.keyword'].value;
                        }

                        String stage = null;
                        if (doc.containsKey('tolqclegacy_assembly_stage.keyword')
                            && doc['tolqclegacy_assembly_stage.keyword'].size() > 0) {
                            stage = doc['tolqclegacy_assembly_stage.keyword'].value;
                        }

                        boolean onSite = doc.containsKey('sts_sample_sts_receive_date_max')
                        && doc['sts_sample_sts_receive_date_max'].size() > 0;

                        boolean releasedToLab =
                        doc.containsKey('sts_sample_sts_tollab_assign_date_min')
                        && doc['sts_sample_sts_tollab_assign_date_min'].size() > 0;

                        boolean noDataYet =
                        (status == null)
                        && (stage == null || stage == '-')
                        && doc.containsKey('mlwh_run_data_mlwh_run_complete_pacbio_min')
                        && doc['mlwh_run_data_mlwh_run_complete_pacbio_min'].size() == 0
                        && doc.containsKey('mlwh_run_data_mlwh_run_complete_hic_min')
                        && doc['mlwh_run_data_mlwh_run_complete_hic_min'].size() == 0;

                        boolean submitted = status == '1 submitted' && stage == 'RELEASED';

                        boolean curation = status == '2 curated' || status == '3 curation';

                        boolean dataComplete = status == '4 data complete';

                        boolean dataIssue = status == '5 data issue';

                        boolean dataGeneration = status == '6 data generation';

                        if (onSite) {
                            emit('a. Species on site');

                            if (!releasedToLab) {
                                emit('b. Not released to lab');
                            } else {
                                if (noDataYet) {
                                    emit('c. No data yet');
                                }
                                if (dataGeneration) {
                                    emit('d. Data generation');
                                }
                                if (dataIssue) {
                                    emit('e. Data issue');
                                }
                                if (dataComplete) {
                                    emit('f. Data complete');
                                }
                                if (curation) {
                                    emit('g. Curation');
                                }
                                if (submitted) {
                                    emit('h. Submitted');
                                }
                            }
                        } else {
                            emit('i. Species not on site');
                        }
                    """
                }
            },
            'calc_specimen_needed_psyche': {
                'type': 'boolean',
                'script': {
                    'source': """
                        boolean isRecollectionRequired = (
                            doc.containsKey('sts_sequencing_material_status.keyword') &&
                            doc['sts_sequencing_material_status.keyword'].size() > 0 &&
                            doc['sts_sequencing_material_status.keyword'].value
                            == 'RECOLLECTION_REQUIRED'
                        );

                        boolean isIncludedProject = (
                            doc.containsKey('goat_long_list.keyword') &&
                            doc['goat_long_list.keyword'].size() > 0 &&
                            ['AG100PEST', 'i5K', 'CBP', 'ERGA-PIL', 'ERGA-BGE', 'ERGA-CH',
                            'ENDEMIXIT'].contains(doc['goat_long_list.keyword'].value)
                        );

                        boolean isChromosome = (
                            doc.containsKey('goat_assembly_level.keyword') &&
                            doc['goat_assembly_level.keyword'].size() > 0 &&
                            doc['goat_assembly_level.keyword'].value == 'Chromosome'
                        );

                        boolean isSpecimensAtSanger = (
                            doc.containsKey('tolid_tolid_count') &&
                            doc['tolid_tolid_count'].size() > 0
                        );

                        boolean isSampleCollected = (
                            doc.containsKey('goat_sample_collected.keyword') &&
                            doc['goat_sample_collected.keyword'].size() > 0 &&
                            ['AG100PEST', 'i5K', 'CNGB', 'CANBP', 'CBP',
                            'ERGA-PIL', 'ERGA-BGE', 'ERGA-CH','ENDEMIXIT'].contains(
                            doc['goat_sample_collected.keyword'].value)
                        );

                        emit(
                            isRecollectionRequired || (!isIncludedProject
                            && !isChromosome && !isSpecimensAtSanger && !isSampleCollected)
                        );
                    """
                }
            },
            'calc_recollection_needed': {
                'type': 'boolean',
                'script': {
                    'source': """
                        boolean isTopUpCountZero = (
                            doc.containsKey
                            ('calc_topup_required_tolid_calc_topup_required_value_count')
                            && doc
                            ['calc_topup_required_tolid_calc_topup_required_value_count']
                            .size() > 0 && doc
                            ['calc_topup_required_tolid_calc_topup_required_value_count']
                            .value == 0
                        );

                        boolean isIndividualExhaustedCountZero = (
                            doc.containsKey
                            ('calc_individual_exhausted_tolid_calc_individual_exhausted_value_count')
                            && doc
                            ['calc_individual_exhausted_tolid_calc_individual_exhausted_value_count']
                            .size() > 0 && doc
                            ['calc_individual_exhausted_tolid_calc_individual_exhausted_value_count']
                            .value == 0
                        );

                        boolean isIndividualNovel = (
                            isTopUpCountZero && isIndividualExhaustedCountZero
                        );

                        boolean isAllTopUpRequired = (
                            doc.containsKey('calc_tolid_calc_topup_required_min') &&
                            doc['calc_tolid_calc_topup_required_min'].size() > 0 &&
                            doc['calc_tolid_calc_topup_required_min'].value == 1
                        );

                        boolean isAllIndividualsExhausted = (
                            doc.containsKey('calc_tolid_calc_individual_exhausted_min') &&
                            doc['calc_tolid_calc_individual_exhausted_min'].size() > 0 &&
                            doc['calc_tolid_calc_individual_exhausted_min'].value == 1
                        );

                        boolean isRecollectionNeeded = (
                            doc.containsKey
                            ('calc_individual_exhausted_tolid_calc_individual_exhausted_value_count')
                            && doc.containsKey('tolid_tolid_count') && doc
                            ['calc_individual_exhausted_tolid_calc_individual_exhausted_value_count']
                            .size() > 0 && doc['tolid_tolid_count'].size() > 0 && doc
                            ['calc_individual_exhausted_tolid_calc_individual_exhausted_value_count']
                            .value - doc['tolid_tolid_count'].value == 0
                        );

                        emit(
                            !isIndividualNovel && isAllTopUpRequired
                            && isAllIndividualsExhausted && isRecollectionNeeded
                        );
                    """
                }
            },
            'calc_species_out_for_recollection': {
                'type': 'boolean',
                'script': {
                    'source': """
                        if (doc.containsKey('portaldb_date_marked_for_recollection') &&
                            doc['portaldb_date_marked_for_recollection'].size() > 0) {
                            if (doc.containsKey('sts_sample_sts_submit_date_max') &&
                                doc['sts_sample_sts_submit_date_max'].size() > 0 &&
                                doc['portaldb_date_marked_for_recollection'].value.isBefore(
                                doc['sts_sample_sts_submit_date_max'].value)) {
                                    emit (false);
                            } else {
                                emit (true);
                            }
                        } else {
                            emit (false);
                        }
                    """
                }
            },
        },
        'specimen': {
            'calc_coverage_post_run': RuntimeFields.math(
                'mlwh_run_data_mlwh_hifi_read_bases_sum',
                'sts_estimated_genome_size',
                operation='/'
            )
        },
        'tolid': {
            'calc_coverage': RuntimeFields.math('mlwh_run_data_mlwh_hifi_read_bases_sum',
                                                'tolid_species.sts_genome_size',
                                                operation='/'),
            'calc_ongoing_submissions': RuntimeFields.math(
                'benchling_pacbio_sequencing_request_count',
                'benchling_pacbio_completed_sequencing_request_count',
                operation='-'),
            'calc_coverage_met': {
                'type': 'boolean',
                'script': {
                    'source': """
                        if (doc.containsKey('mlwh_run_data_mlwh_hifi_read_bases_sum') &&
                        doc.containsKey('tolid_species.sts_genome_size') &&
                        doc.containsKey('sts_sample_sts_target_coverage_max') &&
                        doc['mlwh_run_data_mlwh_hifi_read_bases_sum'].size() > 0 &&
                        doc['tolid_species.sts_genome_size'].size() > 0 &&
                        doc['sts_sample_sts_target_coverage_max'].size() > 0) {
                            emit(doc['mlwh_run_data_mlwh_hifi_read_bases_sum'].value /
                                doc['tolid_species.sts_genome_size'].value >=
                                doc['sts_sample_sts_target_coverage_max'].value)
                        }
                        else {
                            emit(false)
                        }
                    """
                }
            },
            'calc_topup_required': {
                'type': 'boolean',
                'script': {
                    'source': """
                        boolean isTotalSubmissionsGreaterThanZero = (
                            doc.containsKey('benchling_pacbio_sequencing_request_count') &&
                            doc['benchling_pacbio_sequencing_request_count'].size() > 0 &&
                            doc['benchling_pacbio_sequencing_request_count'].value > 0
                        );

                        boolean isOngoingSubmissionsEqualZero = (
                            doc.containsKey('benchling_pacbio_sequencing_request_count') &&
                            doc.containsKey('benchling_pacbio_completed_sequencing_request_count')
                            && doc['benchling_pacbio_sequencing_request_count'].size() > 0 &&
                            doc['benchling_pacbio_completed_sequencing_request_count'].size() > 0
                            && (doc['benchling_pacbio_sequencing_request_count'].value -
                            doc['benchling_pacbio_completed_sequencing_request_count'].value == 0)
                        );

                        boolean isTargetCoverageMet = (
                            doc.containsKey('mlwh_run_data_mlwh_hifi_read_bases_sum') &&
                            doc.containsKey('tolid_species.sts_genome_size') &&
                            doc.containsKey('sts_sample_sts_target_coverage_max') &&
                            doc['mlwh_run_data_mlwh_hifi_read_bases_sum'].size() > 0 &&
                            doc['tolid_species.sts_genome_size'].size() > 0 &&
                            doc['sts_sample_sts_target_coverage_max'].size() > 0 &&
                            (doc['mlwh_run_data_mlwh_hifi_read_bases_sum'].value /
                            doc['tolid_species.sts_genome_size'].value >=
                            doc['sts_sample_sts_target_coverage_max'].value)
                            );

                        boolean isSpecimenNotAtSequencingStage = false;

                            if (doc.containsKey('informatics_status_summary.keyword') &&
                                doc['informatics_status_summary.keyword'].size() > 0) {

                                for (String value : doc['informatics_status_summary.keyword']) {
                                    if (value == '1 submitted' ||
                                        value == '2 curated' ||
                                        value == '3 curation' ||
                                        value == '4 data complete' ||
                                        value == '7 ignore') {
                                        isSpecimenNotAtSequencingStage = true;
                                        break;
                                    }
                                }
                            }
                        emit(
                            isTotalSubmissionsGreaterThanZero &&
                            isOngoingSubmissionsEqualZero &&
                            !isTargetCoverageMet &&
                            !isSpecimenNotAtSequencingStage
                        );
                    """
                }
            },
            'calc_tolid_actionable': {
                'type': 'boolean',
                'script': {
                    'source': """
                        boolean isTolidNotBeenActioned = (
                            !doc.containsKey('portaldb_date_topup_actioned') ||
                            doc['portaldb_date_topup_actioned'].size() == 0 ||
                            doc['portaldb_date_topup_actioned'].value == null
                        );

                        boolean isTolidActionedBeforeLastRun = false;
                        if (doc.containsKey('mlwh_run_data_mlwh_run_complete_max') &&
                            doc.containsKey('portaldb_date_topup_actioned') &&
                            doc['mlwh_run_data_mlwh_run_complete_max'].size() > 0 &&
                            doc['portaldb_date_topup_actioned'].size() > 0) {

                            isTolidActionedBeforeLastRun =
                                doc['portaldb_date_topup_actioned'].value.isBefore(
                                doc['mlwh_run_data_mlwh_run_complete_max'].value);
                        }

                        emit(isTolidNotBeenActioned || isTolidActionedBeforeLastRun);
                    """
                }
            },
            'calc_individual_exhausted': {
                'type': 'boolean',
                'script': {
                    'source': """
                        boolean isIndividualExhausted = (
                            doc.containsKey(
                            'calc_sequencing_request_calc_mlwh_volume_remaining_max'
                            ) && doc.containsKey(
                            'calc_extraction_calc_benchling_volume_ul_dna_max'
                            ) && doc.containsKey(
                            'calc_tissue_prep_calc_benchling_weight_mg_max'
                            ) && doc.containsKey(
                            'calc_sample_calc_benchling_remaining_weight_max'
                            ) && doc.containsKey(
                            'benchling_sample_count'
                            ) && doc.containsKey(
                            'sts_sample_count'
                            ) && doc[
                            'calc_sequencing_request_calc_mlwh_volume_remaining_max'
                            ].size() > 0 && doc[
                            'calc_extraction_calc_benchling_volume_ul_dna_max'
                            ].size() > 0 && doc[
                            'calc_tissue_prep_calc_benchling_weight_mg_max'
                            ].size() > 0 && doc[
                            'calc_sample_calc_benchling_remaining_weight_max'
                            ].size() > 0 && doc[
                            'benchling_sample_count'
                            ].size() > 0 && doc[
                            'sts_sample_count'
                            ].size() > 0 &&
                            doc[
                            'calc_sequencing_request_calc_mlwh_volume_remaining_max'
                            ].value <= 0 &&
                            doc['calc_extraction_calc_benchling_volume_ul_dna_max'
                            ].value <= 0 && doc[
                            'calc_tissue_prep_calc_benchling_weight_mg_max'
                            ].value <= 0 && doc[
                            'calc_sample_calc_benchling_remaining_weight_max'
                            ].value <= 0 && doc[
                            'benchling_sample_count'
                            ].value == doc[
                            'sts_sample_count'
                            ].value
                            );

                        emit(
                            isIndividualExhausted
                        );
                    """
                }
            },
            'calc_individual_available': {
                'type': 'boolean',
                'script': {
                    'source': """
                        boolean isThereAtLeastAnotherIndividual = (
                            doc.containsKey('tolid_species.tolid_tolid_count') &&
                            doc['tolid_species.tolid_tolid_count'].size() > 0 &&
                            doc['tolid_species.tolid_tolid_count'].value > 1
                        );

                         boolean isAtLeastOneIndividualExhausted = (
                            doc.containsKey
                            ('tolid_species.calc_tolid_calc_individual_exhausted_max') &&
                            doc['tolid_species.calc_tolid_calc_individual_exhausted_max']
                            .size() > 0 &&
                            doc['tolid_species.calc_tolid_calc_individual_exhausted_max']
                            .value == 1

                        );

                        boolean isSpeciesTopUpEqualsIndividualExhausted = (
                            doc.containsKey
                            ('tolid_species.calc_topup_required_tolid_calc_topup_required_value_count')
                            && doc.containsKey
                            ('tolid_species.calc_individual_exhausted_tolid_calc_individual_exhausted_value_count')
                            && doc
                            ['tolid_species.calc_topup_required_tolid_calc_topup_required_value_count']
                            .size() > 0 && doc
                            ['tolid_species.calc_individual_exhausted_tolid_calc_individual_exhausted_value_count']
                            .size() > 0 && (doc
                            ['tolid_species.calc_topup_required_tolid_calc_topup_required_value_count']
                            .value - doc
                            ['tolid_species.calc_individual_exhausted_tolid_calc_individual_exhausted_value_count']
                            .value == 0)
                        );

                        emit(
                            isThereAtLeastAnotherIndividual &&
                            isAtLeastOneIndividualExhausted &&
                            isSpeciesTopUpEqualsIndividualExhausted
                        );
                    """
                }
            },
        },
        'sample': {
            'calc_biospecimen_id': RuntimeFields.coalesce([
                'sts_sample_same_as',
                'sts_biospecimen_accession',
                'sts_sample_symbiont_of'
            ]),
            'calc_sts_export_eligible': {
                'type': 'boolean',
                'script': {
                    'source': """
                        String biospecimenId = null;
                        List biospecimenFields = [
                            'sts_sample_same_as.keyword',
                            'sts_biospecimen_accession.keyword',
                            'sts_sample_symbiont_of.keyword'
                        ];

                        for (String field : biospecimenFields) {
                            if (doc.containsKey(field) &&
                                doc[field].size() > 0) {
                                biospecimenId = doc[field].value;
                                break;
                            }
                        }

                        List requiredFields = [
                            'sts_tubeid.keyword',
                            'sts_species.sts_scientific_name.keyword',
                            'sts_species.id.keyword',
                            'sts_specimen.id.keyword',
                            'sts_organism_part.keyword',
                            'sts_lifestage.keyword',
                            'sts_preservation_approach.keyword',
                            'sts_sex.keyword',
                            'sts_tissue_size.keyword',
                            'sts_programme.keyword',
                            'sts_biosample_accession.keyword',
                            'sts_submit_date',
                            'sts_sampleset.id.keyword',
                            'sts_project.keyword',
                            'sts_species.sts_taxon_group.keyword',
                            'sts_species.sts_genome_size',
                            'sts_rackid.keyword',
                            'sts_pos_in_rack.keyword',
                            'sts_labwhere_parentage.keyword',
                            'sts_labwhere_name.keyword',
                            'sts_cost_code.keyword',
                            'sts_sequencescape_study_id.keyword'
                        ];

                        boolean basicFieldsExist = true;
                        for (String field : requiredFields) {
                            boolean fieldHasValue = false;

                            if (doc.containsKey(field)) {
                                if (field.endsWith(".keyword")) {
                                    fieldHasValue = doc[field].size() > 0;
                                } else {
                                    fieldHasValue = doc[field] != null;
                                }
                            }

                            if (!fieldHasValue) {
                                basicFieldsExist = false;
                                break;
                            }
                        }

                        boolean biospecimenIdExists = (biospecimenId != null);
                        emit(basicFieldsExist && biospecimenIdExists);
                    """
                }
            },
            'calc_benchling_remaining_weight': {
                'type': 'double',
                'script': {
                    'source': """
                                            if (doc.containsKey('portaldb_date_abandoned') &&
                            doc['portaldb_date_abandoned'].size() > 0) {
                            emit(0.0);
                        } else if (doc.containsKey('benchling_remaining_weight') &&
                            doc['benchling_remaining_weight'].size() > 0) {
                            def value = doc['benchling_remaining_weight'].value;
                            if (!Double.isNaN(value)) {
                                emit(value);
                            } else {
                                emit(0.5);
                            }
                        } else {
                            emit(0.5);
                        }
                    """
                }
            },
            'calc_sample_abandoned_in_sts': {
                'type': 'boolean',
                'script': {
                    'source': """
                        boolean isSampleAbandoned = doc.containsKey('portaldb_date_abandoned') &&
                            doc['portaldb_date_abandoned'].size() > 0;

                        boolean isSampleInBenchling = doc.containsKey('sts_eln_id.keyword') &&
                            doc['sts_eln_id.keyword'].size() > 0;

                        if (isSampleAbandoned && !isSampleInBenchling) {
                            emit(true);
                        } else {
                            emit(false);
                        }
                    """
                }
            },
        },
        'sampleset': {
            'calc_tat': RuntimeFields.date_interval('sts_submit_date',
                                                    'sts_sample_sts_receive_date_min',
                                                    'days')
        },
        'manifest': {
            'calc_tat': RuntimeFields.date_interval('sts_submit_date',
                                                    'sts_receive_date',
                                                    'days')
        },
        'sequencing_request': {
            'calc_existing_library_oplc': {
                'type': 'double',
                'script': {
                    'source': """
                      if (doc.containsKey('mlwh_insert_size')
                      && doc['mlwh_insert_size'].size() > 0
                      && doc.containsKey('mlwh_concentration')
                      && doc['mlwh_concentration'].size() > 0
                      && doc.containsKey('mlwh_volume_remaining')
                      && doc['mlwh_volume_remaining'].size() > 0){
                        emit((doc['mlwh_concentration']
                        .value/(doc['mlwh_insert_size'].value * 660))
                        * doc['mlwh_volume_remaining'].value * 7500000)
                }
                """
                }
            },
            'calc_mlwh_volume_remaining': {
                'type': 'double',
                'script': {
                    'source': """
                    if (doc.containsKey('portaldb_date_abandoned') &&
                        doc['portaldb_date_abandoned'].size() > 0) {
                        emit(0.0);
                    } else if (doc.containsKey('mlwh_volume_remaining') &&
                        doc['mlwh_volume_remaining'].size() > 0) {
                        def value = doc['mlwh_volume_remaining'].value;
                        if (!Double.isNaN(value)) {
                emit(value);
            } else {
                emit(0.5);
            }
        } else {
            emit(0.5);
        }
                """
                }
            },
        },
        'extraction': {
            'calc_dna_volume_remaining': RuntimeFields.coalesce([
                'mlwh_volume_si_value',
                'benchling_volume_ul',
            ], return_type='double'
            ),
            'calc_benchling_volume_ul': {
                'type': 'double',
                'script': {
                    'source': """
                    if (doc.containsKey('portaldb_date_abandoned') &&
                        doc['portaldb_date_abandoned'].size() > 0) {
                        emit(0.0);
                    } else if (doc.containsKey('benchling_volume_ul') &&
                        doc['benchling_volume_ul'].size() > 0) {
                        def value = doc['benchling_volume_ul'].value;
                        if (!Double.isNaN(value)) {
                emit(value);
            } else {
                emit(0.5);
            }
        } else {
            emit(0.5);
        }
                """
                }
            },
        },
        'tissue_prep': {
            'calc_benchling_weight_mg': {
                'type': 'double',
                'script': {
                    'source': """
                    if (doc.containsKey('portaldb_date_abandoned') &&
                        doc['portaldb_date_abandoned'].size() > 0) {
                        emit(0.0);
                    } else if (doc.containsKey('benchling_weight_mg') &&
                        doc['benchling_weight_mg'].size() > 0) {
                        def value = doc['benchling_weight_mg'].value;
                        if (!Double.isNaN(value)) {
                emit(value);
            } else {
                emit(0.5);
            }
        } else {
            emit(0.5);
        }
                """
                }
            }
        },
    }

    amd = data_source_attribute_metadata(
        portal_attributes()
    )

    elastic = ElasticDataSource({
        'uri': os.getenv('ELASTIC_URI'),
        'user': os.getenv('ELASTIC_USER'),
        'password': os.getenv('ELASTIC_PASSWORD'),
        'index_prefix': os.getenv('ELASTIC_INDEX_PREFIX'),
        'relationship_cfg': relationship_config,
        'runtime_fields': runtime_fields},
        attribute_metadata=amd)
    core_data_object(elastic)
    return elastic



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT


def convert_tolid_specimen_to_data_object(specimens, target_datasource):
    CoreDataObject = target_datasource.data_object_factory  # noqa N806
    for s in specimens:
        ret = CoreDataObject('specimen', {
            'specimen_id': s.specimen_id,
            'tolid': s.tolid,
            'taxon_id': s.species.id
        })
        yield ret



# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import Any, Callable, Iterable, Optional, Type

from flask import make_response

from .auth import AuthInspector
from .misc import (
    AggregationBody,
    AggregationParameters,
    GroupStatsParameters,
    JsonApiRequestBody,
    ListGetParamaters,
    StatsParameters
)
from ..api_client.exception import (
    ObjectNotFoundByIdException,
    RecursiveRelationNotFoundException,
    UninheritedOperationError,
    UnsupportedOpertionError
)
from ..api_client.view import ResponseDict, View
from ..core import DataObject, OperableDataSource
from ..core.datasource_filter import (
    AndFilter,
    DataSourceFilter
)
from ..core.operator import (
    Aggregator,
    Counter,
    Cursor,
    Deleter,
    DetailGetter,
    GroupStatter,
    Inserter,
    ListGetter,
    Operator,
    OperatorMethod,
    PageGetter,
    Relational,
    ReturnMode,
    Updater,
    Upserter
)
from ..core.operator.updater import DataObjectUpdate
from ..excel import convert_data_objects_to_excel


EmptySuccessResponse = dict[str, bool]


def __is_supported(
    operator_class: Type[Operator],
    operator_method: str,
    data_source: OperableDataSource,
) -> bool:
    """
    Returns `True` if the given `DataSource` instance implements
    the specified `Operator` class.

    If it doesn't, it first checks if the operator method is
    defined on the `DataSource` instance, as this implies an
    inheritance error, and raises an `UninheritedOperationError`
    if so.

    Finally, otherwise, returns `False`
    """
    if isinstance(data_source, operator_class):
        return True
    if hasattr(data_source, operator_method):
        raise UninheritedOperationError(
            data_source,
            operator_class,
            operator_method
        )
    return False


def validate(
    operator_class: Type[Operator],
    object_method_name: str,
    operator_method: OperatorMethod
) -> Callable:
    """
    Validates:
    - a Controller method's corresponding operation is supported by its DataSource:
        - an UninheritedOperationError is raised if the method is implemented
          but the mixin ABC is not inherited from.
        - otherwise an UnsupportedOpertionError is raised.
    """
    def decorator(method: Callable) -> Callable:
        def wrapper(controller: Controller, object_type: str, *args, **kwargs) -> Any:
            if not __is_supported(
                operator_class,
                object_method_name,
                controller.data_source
            ):
                raise UnsupportedOpertionError(
                    object_type,
                    str(operator_method)
                )
            ext_and = controller.inspect_auth(object_type, operator_method)
            return method(
                controller,
                object_type,
                *args,
                ext_and=ext_and,
                **kwargs,
            )
        return wrapper
    return decorator


class Controller:
    """
    An MVC-esque Controller class, that fulfills requests.
    """

    def __init__(
        self,
        data_source: OperableDataSource,
        view: View,
        auth_inspector: Optional[AuthInspector] = None
    ) -> None:

        self.__data_source = data_source
        self.__view = view
        self.__inspector = auth_inspector

    @property
    def data_source(self) -> OperableDataSource:
        return self.__data_source

    def inspect_auth(
        self,
        object_type: str,
        operation: OperatorMethod
    ) -> Optional[AndFilter]:

        if self.__inspector is not None:
            return self.__inspector(object_type, operation)

    @validate(DetailGetter, 'get_by_id', OperatorMethod.DETAIL)
    def get_detail(self, object_type: str, object_id: str, **kwargs) -> ResponseDict:
        """
        Gets an individual object of specified type and id
        """
        data_object = self.__get_detail_object(object_type, object_id)
        return self.__view.dump(data_object)

    @validate(PageGetter, 'get_list_page', OperatorMethod.PAGE)
    def get_list(
        self,
        object_type: str,
        query_args: ListGetParamaters,
        ext_and: Optional[AndFilter] = None
    ) -> ResponseDict:
        """
        Gets a page of list results of specified type.
        """
        page_number = self.__get_page_number_or_1(query_args)
        data_objects, total = self.__data_source.get_list_page(
            object_type,
            page_number,
            page_size=query_args.page_size,
            object_filters=self.__combine_filters(
                query_args.filter,
                ext_and
            ),
            sort_by=query_args.sort_by,
            requested_fields=query_args.requested_fields
        )
        document_meta = {
            'total': total,
            'types': self.__data_source.get_attribute_types(object_type)
        }
        return self.__view.dump_bulk(data_objects, document_meta=document_meta)

    @validate(ListGetter, 'post_list_export', OperatorMethod.EXPORT)
    def post_list_export(
        self,
        object_type: str,
        query_args: ListGetParamaters,
        body: JsonApiRequestBody,
        ext_and: Optional[AndFilter],
    ) -> ResponseDict:
        """
        Gets all the list results of the specified type.
        """

        page_number = self.__get_page_number_or_1(query_args)
        data_objects, _ = self.__data_source.get_list_page(
            object_type,
            page_number,
            page_size=query_args.page_size,
            object_filters=self.__combine_filters(
                query_args.filter,
                ext_and
            ),
            sort_by=query_args.sort_by
        )

        output_stream = convert_data_objects_to_excel(
            self.data_source,
            object_type,
            data_objects,
            body.data,
            'Sheet1'
        )
        response = make_response(output_stream.getvalue())
        response.headers['Content-Disposition'] = 'attachment; filename=download_table.xlsx'
        response.headers['Content-type'] = 'application/vnd.ms-excel'
        return response

    @validate(Counter, 'get_count', OperatorMethod.COUNT)
    def get_count(
        self,
        object_type: str,
        query_args: ListGetParamaters,
        ext_and: Optional[AndFilter],
    ) -> ResponseDict:
        """
        Gets a count of specified object type, respecting filters.
        """
        total = self.__data_source.get_count(
            object_type,
            object_filters=self.__combine_filters(
                query_args.filter,
                ext_and
            ),
        )
        document_meta = {
            'total': total
        }
        return self.__view.dump_bulk([], document_meta=document_meta)

    @validate(Counter, 'get_stats', OperatorMethod.STATS)
    def get_stats(
        self,
        object_type: str,
        query_args: StatsParameters,
        ext_and: Optional[AndFilter],
    ) -> ResponseDict:
        """
        Gets stats of specified object type, respecting filters.
        """
        stats = self.__data_source.get_stats(
            object_type,
            stats=query_args.stats,
            stats_fields=query_args.stats_fields,
            object_filters=self.__combine_filters(
                query_args.filter,
                ext_and
            ),
        )
        document_meta = {**stats, 'type': object_type}
        return self.__view.dump_bulk([], document_meta=document_meta)

    @validate(GroupStatter, 'get_group_stats', OperatorMethod.GROUP_STATS)
    def get_group_stats(
        self,
        object_type: str,
        query_args: GroupStatsParameters,
        ext_and: Optional[AndFilter],
    ) -> ResponseDict:
        """
        Gets stats of specified object type, respecting filters.
        """
        stats = self.__data_source.get_group_stats(
            object_type,
            query_args.group_by,
            stats=query_args.stats,
            stats_fields=query_args.stats_fields,
            object_filters=self.__combine_filters(
                query_args.filter,
                ext_and
            ),
        )
        document_meta = {
            'stats': list(stats),
            'type': object_type
        }
        return self.__view.dump_bulk([], document_meta=document_meta)

    @validate(Deleter, 'delete', OperatorMethod.DELETE)
    def delete_detail(
        self,
        object_type: str,
        object_id: str,
        **kwargs
    ) -> EmptySuccessResponse:
        """Deletes the `DataObject` of specified type and id"""

        self.data_source.delete(object_type, [object_id])
        return {'success': True}

    @validate(Updater, 'update', OperatorMethod.UPDATE)
    def patch_list(
        self,
        object_type: str,
        updates: Iterable[DataObjectUpdate],
        **kwargs
    ) -> EmptySuccessResponse:
        """
        Updates the objects (all of same type) using the given
        `Iterable` of ID:update-dict pairs
        """

        self.data_source.update(object_type, updates)
        return {'success': True}

    @validate(Inserter, 'insert', OperatorMethod.INSERT)
    def post_inserts(
        self,
        object_type: str,
        objects: Iterable[DataObject],
        **kwargs
    ) -> EmptySuccessResponse:
        """Inserts the given objects of specified type"""

        returned = self.data_source.insert(object_type, objects)
        if self.data_source.return_mode[object_type] == ReturnMode.POPULATED:
            return self.__view.dump_bulk(returned)
        else:
            return {'success': True}

    @validate(Upserter, 'post_upserts', OperatorMethod.UPSERT)
    def post_upserts(
        self,
        object_type: str,
        objects: Iterable[DataObject],
        **kwargs
    ) -> EmptySuccessResponse:
        """Upserts the given objects of specified type"""

        returned = self.data_source.upsert(object_type, objects)
        if self.data_source.return_mode[object_type] == ReturnMode.POPULATED:
            return self.__view.dump_bulk(returned)
        else:
            return {'success': True}

    @validate(Aggregator, 'get_aggregations', OperatorMethod.AGGREGATE)
    def post_aggregations(
        self,
        object_type: str,
        query_args: AggregationParameters,
        body: AggregationBody,
        ext_and: Optional[AndFilter],
    ) -> ResponseDict:
        """
        Gets an aggregation on the specified object_type.
        """
        aggregation_results = self.__data_source.get_aggregations(
            object_type,
            object_filters=self.__combine_filters(
                query_args.filter,
                ext_and
            ),
            aggregations=body.aggs
        )
        document_meta = {
            'aggregations': aggregation_results,
            'types': self.__data_source.get_attribute_types(object_type)
        }
        return self.__view.dump_bulk([], document_meta=document_meta)

    @validate(
        Cursor,
        'get_cursor_page',
        OperatorMethod.CURSOR
    )
    def get_cursor_page(
        self,
        object_type: str,
        query_args: ListGetParamaters,
        search_after: list[str] | None,
        ext_and: Optional[AndFilter]
    ) -> ResponseDict:

        data_objects, new_search_after = self.data_source.get_cursor_page(
            object_type,
            query_args.page_size,
            self.__combine_filters(
                query_args.filter,
                ext_and
            ),
            search_after
        )
        meta = {'search_after': new_search_after}

        return self.__view.dump_bulk(
            data_objects,
            document_meta=meta
        )

    @validate(
        Relational,
        'get_recursive_relation',
        OperatorMethod.TO_ONE
    )
    def get_recursive_relation(
        self,
        data_object: DataObject,
        relationship_hops: list[str],
        **kwargs
    ) -> ResponseDict:
        """
        Gets a nested to-one relation by recursive hops
        """

        related_object = self.__get_to_one_relation(
            data_object,
            relationship_hops
        )
        if related_object is None:
            raise RecursiveRelationNotFoundException()
        return self.__view.dump(related_object)

    @validate(
        Relational,
        'get_to_many_relations_page',
        OperatorMethod.TO_MANY
    )
    def get_many_relations_page(
        self,
        data_object: DataObject,
        relationship_name: str,
        query_args: ListGetParamaters,
        **kwargs
    ) -> ResponseDict:
        """Gets a page of to-many relation results"""

        page = self.data_source.get_to_many_relations_page(
            data_object,
            relationship_name,
            query_args.page,
            query_args.page_size
        )
        return self.__view.dump_bulk(page)

    def __combine_filters(
        self,
        object_filters: Optional[DataSourceFilter],
        ext_and: Optional[AndFilter]
    ) -> Optional[DataSourceFilter]:

        if ext_and is None:
            return object_filters
        else:
            if object_filters is None:
                return DataSourceFilter(
                    and_=ext_and
                )
            elif object_filters.and_ is None:
                object_filters.and_ = ext_and
            else:
                object_filters.and_ |= ext_and

            return object_filters

    def __get_detail_object(self, object_type: str, object_id: str) -> DataObject:
        data_objects = list(self.__data_source.get_by_id(object_type, [object_id]))
        if len(data_objects) == 0 or data_objects[0] is None:
            raise ObjectNotFoundByIdException(object_type, object_id)
        return data_objects[0]

    def __get_page_number_or_1(self, query_args: ListGetParamaters) -> int:
        page_number = query_args.page
        if page_number is None:
            return 1
        return page_number

    def __get_to_one_relation(
        self,
        source: DataObject,
        relationship_hops: list[str]
    ) -> Optional[DataObject]:

        self.__data_source.validate_to_one_recurse(
            source.type,
            relationship_hops
        )
        return self.__data_source.get_recursive_relation(
            source,
            relationship_hops
        )



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT


def convert_elastic_samples_to_barcoding_run_data(samples, target_datasource):
    for sample in samples:
        CoreDataObject = target_datasource.data_object_factory  # noqa N806
        ret = CoreDataObject(
            'barcoding_run_data',
            attributes={
                'col_date': sample.col_date,
                'taxon_id': sample.id
            }
        )
        yield ret



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from collections.abc import Iterable as IterableABC
from dataclasses import dataclass
from typing import Any, Iterator

from sqlalchemy import ForeignKey
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import (
    Mapped,
    declared_attr,
    mapped_column,
    relationship
)

from ..model import Model


@dataclass(frozen=True, kw_only=True)
class BoardModels(IterableABC[type[Model]]):
    """
    Contains the needed models for dynamic, user-configurable
    dashboards.

    Additionally contains `_user_mixin`, from which developers
    should inherit their `User` class from `tol.sql.auth`.
    """

    component: type[Model]
    component_zone: type[Model]
    zone: type[Model]
    zone_view: type[Model]
    view: type[Model]
    view_board: type[Model]
    board: type[Model]

    _user_mixin: type[Any]

    def __iter__(self) -> Iterator[type[Model]]:
        """
        Returns in order they should be deleted
        """

        return iter(
            [
                self.component_zone,
                self.component,
                self.zone_view,
                self.zone,
                self.view_board,
                self.view,
                self.board
            ]
        )


def create_board_models(
    base_model_class: type[Model],
    user_model_class_name: str = 'User',
    user_table_name: str = 'user'
) -> BoardModels:
    """
    Creates all needed models (and joining tables) for
    user-configurable dashboards.

    Returns a `BoardModels` instance that functions like an
    `Iterable`.
    """

    class Component(base_model_class):
        __tablename__ = 'component'

        id: Mapped[str] = mapped_column(  # noqa A003
            primary_key=True
        )

        title: Mapped[str] = mapped_column(nullable=False)
        object_type: Mapped[str] = mapped_column(nullable=False)
        base_url: Mapped[str] = mapped_column(nullable=True)
        component_type: Mapped[str] = mapped_column(nullable=False)
        widget_type: Mapped[str] = mapped_column(nullable=False)
        config = mapped_column(type_=JSONB(), nullable=False)
        filter = mapped_column(  # noqa A003
            type_=JSONB(),
            nullable=False,
            default={},
            server_default='{}'  # noqa P103
        )
        filter_pass_through: Mapped[bool] = mapped_column(nullable=False)

        user_id: Mapped[int] = mapped_column(
            ForeignKey(f'{user_table_name}.id'),
            nullable=False
        )
        user = relationship(
            user_model_class_name,
            back_populates='components',
            foreign_keys=[user_id]
        )

        component_zones: Mapped[list[ComponentZone]] = relationship(
            back_populates='component'
        )

    class ComponentZone(base_model_class):
        __tablename__ = 'component_zone'

        id: Mapped[int] = mapped_column(  # noqa A003
            primary_key=True,
            autoincrement=True
        )

        order: Mapped[int] = mapped_column(nullable=False)

        component_id: Mapped[str] = mapped_column(
            ForeignKey('component.id'),
            nullable=False
        )
        component = relationship(
            'Component',
            back_populates='component_zones',
            foreign_keys=[component_id]
        )

        zone_id: Mapped[str] = mapped_column(
            ForeignKey('zone.id'),
            nullable=False
        )
        zone = relationship(
            'Zone',
            back_populates='component_zones',
            foreign_keys=[zone_id]
        )

    class Zone(base_model_class):
        __tablename__ = 'zone'

        id: Mapped[str] = mapped_column(  # noqa A003
            primary_key=True
        )

        title: Mapped[str] = mapped_column(nullable=False)
        object_type: Mapped[str] = mapped_column(nullable=False)
        base_url: Mapped[str] = mapped_column(nullable=True)
        filter = mapped_column(  # noqa A003
            type_=JSONB(),
            nullable=False,
            default={},
            server_default='{}'  # noqa P103
        )

        component_zones: Mapped[list[ComponentZone]] = relationship(
            back_populates='zone'
        )

        zone_views: Mapped[list[ZoneView]] = relationship(
            back_populates='zone'
        )

        user_id: Mapped[int] = mapped_column(
            ForeignKey(f'{user_table_name}.id'),
            nullable=False
        )
        user = relationship(
            user_model_class_name,
            back_populates='zones',
            foreign_keys=[user_id]
        )

    class ZoneView(base_model_class):
        __tablename__ = 'zone_view'

        id: Mapped[int] = mapped_column(  # noqa A003
            primary_key=True,
            autoincrement=True
        )

        order: Mapped[int] = mapped_column(nullable=False)

        zone_id: Mapped[str] = mapped_column(
            ForeignKey('zone.id'),
            nullable=False
        )
        zone = relationship(
            'Zone',
            back_populates='zone_views',
            foreign_keys=[zone_id]
        )

        view_id: Mapped[str] = mapped_column(
            ForeignKey('view.id'),
            nullable=False
        )
        view = relationship(
            'View',
            back_populates='zone_views',
            foreign_keys=[view_id]
        )

    class View(base_model_class):
        __tablename__ = 'view'

        id: Mapped[str] = mapped_column(  # noqa A003
            primary_key=True
        )

        title: Mapped[str] = mapped_column(nullable=False)
        filter = mapped_column(  # noqa A003
            type_=JSONB(),
            nullable=False,
            default={},
            server_default='{}'  # noqa P103
        )

        zone_views: Mapped[list[ZoneView]] = relationship(
            back_populates='view'
        )

        view_boards: Mapped[list[ViewBoard]] = relationship(
            back_populates='view'
        )

        user_id: Mapped[int] = mapped_column(
            ForeignKey(f'{user_table_name}.id'),
            nullable=False
        )
        user = relationship(
            user_model_class_name,
            back_populates='views',
            foreign_keys=[user_id]
        )

    class ViewBoard(base_model_class):
        __tablename__ = 'view_board'

        id: Mapped[int] = mapped_column(  # noqa A003
            primary_key=True,
            autoincrement=True
        )

        order: Mapped[int] = mapped_column(nullable=False)

        view_id: Mapped[str] = mapped_column(
            ForeignKey('view.id'),
            nullable=False
        )
        view = relationship(
            'View',
            back_populates='view_boards',
            foreign_keys=[view_id]
        )

        board_id: Mapped[str] = mapped_column(
            ForeignKey('board.id'),
            nullable=False
        )
        board = relationship(
            'Board',
            back_populates='view_boards',
            foreign_keys=[board_id]
        )

    class Board(base_model_class):
        __tablename__ = 'board'

        id: Mapped[str] = mapped_column(  # noqa A003
            primary_key=True
        )

        title: Mapped[str] = mapped_column(nullable=False)
        filter = mapped_column(  # noqa A003
            type_=JSONB(),
            nullable=False,
            default={},
            server_default='{}'  # noqa P103
        )

        view_boards: Mapped[list[ViewBoard]] = relationship(
            back_populates='board'
        )

        user_id: Mapped[int] = mapped_column(
            ForeignKey(f'{user_table_name}.id'),
            nullable=False
        )
        user = relationship(
            user_model_class_name,
            back_populates='boards',
            foreign_keys=[user_id]
        )

    class _UserMixin:

        @declared_attr
        def components(self) -> Mapped[list[Component]]:
            return relationship(
                back_populates='user'
            )

        @declared_attr
        def zones(self) -> Mapped[list[Zone]]:
            return relationship(
                back_populates='user'
            )

        @declared_attr
        def views(self) -> Mapped[list[View]]:
            return relationship(
                back_populates='user'
            )

        @declared_attr
        def boards(self) -> Mapped[list[Board]]:
            return relationship(
                back_populates='user'
            )

    return BoardModels(
        component=Component,
        component_zone=ComponentZone,
        zone=Zone,
        zone_view=ZoneView,
        view=View,
        view_board=ViewBoard,
        board=Board,
        _user_mixin=_UserMixin
    )



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Callable, Optional

import flask

from ...core import DataSourceError


class NotAuthenticatedError(DataSourceError):
    """
    Raised when attempting to access `AuthContext`
    properties without a successful authentication event.
    """

    def __init__(self) -> None:
        detail = (
            'No user has authenticated for this request'
        )

        super().__init__(
            title='Unauthorized',
            detail=detail,
            status_code=401
        )


class AuthContext:
    """
    The auth context for a specific request, lasting
    only for its duration.

    Used principally to store the ID for the user making
    a request, if authenticated, but can also be used for
    other things.
    """

    def __init__(self) -> None:
        self.__user_id: Optional[str] = None
        self.__roles: list[str] = []

    @property
    def authenticated(self) -> bool:
        """
        `True` if the user has authenticated succesfully,
        perhaps by using a token.
        """

        return self.__user_id is not None

    @property
    def user_id(self) -> str:
        """A `str` that uniquely identifies a user."""

        self.__assert_authenticated()

        return self.__user_id

    @user_id.setter
    def user_id(self, val: str) -> None:
        self.__user_id = val

    @property
    def roles(self) -> list[str]:
        """
        A `list[str]` of names of roles assigned to
        the authenticated user.
        """

        self.__assert_authenticated()

        return self.__roles

    @roles.setter
    def roles(self, val: list[str]) -> None:
        self.__roles = val

    def __assert_authenticated(self) -> None:
        if not self.authenticated:
            raise NotAuthenticatedError()


CtxGetter = Callable[[], AuthContext]
"""
A callable that fetches the global `AuthContext` instance
"""


def default_ctx_getter() -> AuthContext:
    return flask.g.setdefault(
        'auth_context',
        default=AuthContext()
    )



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import urllib
from collections import ChainMap
from itertools import chain
from typing import Optional

from flask import Blueprint, request

from .auth import AuthInspector
from .auth.error import AuthError
from .controller import Controller
from .misc import (
    AggregationBody,
    AggregationParameters,
    GroupStatsParameters,
    JsonApiRequestBody,
    ListGetParamaters,
    RelataionshipHopsParser,
    StatsParameters
)
from ..api_client.exception import BaseRuntimeException
from ..api_client.parser import DefaultParser
from ..api_client.view import DefaultView
from ..core import DataSource, DataSourceError, OperableDataSource
from ..core.data_source_dict import DataSourceDict
from ..core.operator import Relational
from ..core.operator.operator_config import DefaultOperatorConfig, OperatorConfig


class DataBlueprint(Blueprint):
    """
    A flask Blueprint for dynamically routing DataObject endpoints
    defined in DataSource instances.
    """

    def __init__(
        self,
        url_prefix: str
    ) -> None:

        super().__init__(
            'data_source_handler',
            __name__,
            url_prefix=url_prefix
        )


class ConfigBlueprint(Blueprint):
    """
    A flask `Blueprint`, to be nested under `DataBlueprint`, that
    stores configuration information about the stored data as a
    whole.
    """

    def __init__(self, url_prefix: str) -> None:
        super().__init__(
            'data_source_config',
            __name__,
            url_prefix=url_prefix
        )


class CustomBlueprint(Blueprint):
    """
    A flask Blueprint for custom endpoints needing to use
    DataSources.
    """

    def __init__(
        self,
        url_prefix: str,
        name: str = __name__
    ) -> None:

        super().__init__(
            name,
            __name__,
            url_prefix=url_prefix
        )


def _config_blueprint(
    url_prefix: str,
    data_sources: tuple[OperableDataSource],
    operator_config: OperatorConfig
) -> ConfigBlueprint:
    """
    Returns a `ConfigBlueprint` instance given:

    - a `url_prefix`, on which to serve the endpoints
    - `data_sources`, a `tuple` of `DataSource` instances behind the API
    """

    config_handler = ConfigBlueprint(url_prefix)

    @config_handler.route('/relationships', methods=['GET'])
    def get_relationships():
        relationship_configs = chain(
            *[
                d.relationship_config.items()
                for d in data_sources
                if isinstance(d, Relational)
            ]
        )
        return {
            t: d.to_dict() for t, d in relationship_configs
            if not d.empty
        }

    @config_handler.route('/attribute_types', methods=['GET'])
    def get_attribute_types():
        types_list = [d.attribute_types for d in data_sources]
        chain_map = ChainMap(*types_list)
        return dict(chain_map)

    @config_handler.route('/attribute_metadata', methods=['GET'])
    def get_attribute_metadata():
        types_list = [d.attribute_metadata for d in data_sources]
        chain_map = ChainMap(*types_list)
        return dict(chain_map)

    @config_handler.route('/operations', methods=['GET'])
    def get_operations():
        return operator_config.to_dict()

    @config_handler.get('/write_mode')
    def get_relation_write_mode():
        modes_list = [d.write_mode for d in data_sources]
        chain_map = ChainMap(*modes_list)
        return dict(chain_map)

    @config_handler.get('/return_mode')
    def get_return_mode():
        modes_list = [d.return_mode for d in data_sources]
        chain_map = ChainMap(*modes_list)
        return dict(chain_map)

    return config_handler


def _core_blueprint(
    data_source_dict: dict[str, DataSource],
    url_prefix: str,
    auth_inspector: Optional[AuthInspector] = None
) -> DataBlueprint:
    """
    Creates the "core" blueprint, responsible for managing
    (non-metadata) endpoints of the `DataSource` instances
    in the given `dict`.
    """

    data_handler = DataBlueprint(url_prefix=url_prefix)

    def __new_controller(
        object_type: str,
        requested_fields: list[str] | None = None,
    ) -> Controller:

        hop_limit = None if requested_fields else 1

        data_source = data_source_dict[object_type]
        view = DefaultView(
            prefix=url_prefix,
            include_all_to_ones=True,
            hop_limit=hop_limit,
            requested_fields=requested_fields,
        )
        return Controller(
            data_source,
            view,
            auth_inspector=auth_inspector
        )

    @data_handler.route('/<object_type>/<path:object_id>', methods=['GET'])  # Allow slashes
    def get_detail(*, object_type: str, object_id: str):
        controller = __new_controller(object_type)
        object_id_unencoded = urllib.parse.unquote(object_id)
        return controller.get_detail(object_type, object_id_unencoded)

    @data_handler.route('/<object_type>', methods=['GET'])
    def get_list(*, object_type: str):
        request_args = ListGetParamaters(request.args)
        controller = __new_controller(
            object_type,
            requested_fields=request_args.requested_fields,
        )
        return controller.get_list(object_type, request_args)

    @data_handler.route('/<object_type>:export', methods=['POST'])
    def post_list_export(*, object_type: str):
        controller = __new_controller(object_type)
        request_args = ListGetParamaters(request.args)
        body = JsonApiRequestBody(request.json)
        response = controller.post_list_export(object_type, request_args, body)
        return response, 200, {'Content-Type': 'application/vnd.ms-excel'}

    @data_handler.route('/<object_type>:count', methods=['GET'])
    def get_count(*, object_type: str):
        controller = __new_controller(object_type)
        request_args = ListGetParamaters(request.args)
        return controller.get_count(object_type, request_args)

    @data_handler.route('/<object_type>:stats', methods=['GET'])
    def get_stats(*, object_type: str):
        controller = __new_controller(object_type)
        request_args = StatsParameters(request.args)
        return controller.get_stats(object_type, request_args)

    @data_handler.get('/<object_type>:group-stats')
    def get_group_stats(*, object_type: str):
        controller = __new_controller(object_type)
        request_args = GroupStatsParameters(request.args)
        return controller.get_group_stats(object_type, request_args)

    @data_handler.route('/<object_type>/<path:object_id>', methods=['DELETE'])
    def delete_detail(*, object_type: str, object_id: str):
        controller = __new_controller(object_type)
        object_id_unencoded = urllib.parse.unquote(object_id)
        return controller.delete_detail(object_type, object_id_unencoded)

    @data_handler.route('/<object_type>', methods=['PATCH'])
    def patch_list(*, object_type: str):
        controller = __new_controller(object_type)
        request_body = JsonApiRequestBody(request.json)
        return controller.patch_list(object_type, request_body.data)

    @data_handler.post('/<object_type>:insert')
    def post_inserts(*, object_type: str):
        controller = __new_controller(object_type)
        request_body = JsonApiRequestBody(request.json)
        parser = DefaultParser(data_source_dict)
        objects = parser.parse_iterable(request_body.data)
        return controller.post_inserts(object_type, objects)

    @data_handler.route('/<object_type>:upsert', methods=['POST'])
    def post_upserts(*, object_type: str):
        controller = __new_controller(object_type)
        request_body = JsonApiRequestBody(request.json)
        parser = DefaultParser(data_source_dict)
        objects = parser.parse_iterable(request_body.data)
        return controller.post_upserts(object_type, objects)

    @data_handler.route('/<object_type>:aggregations', methods=['POST'])
    def get_aggregations(*, object_type: str):
        controller = __new_controller(object_type)
        request_args = AggregationParameters(request.args)
        body = AggregationBody(request.json)
        return controller.post_aggregations(object_type, request_args, body)

    @data_handler.post('/<object_type>:cursor')
    def get_cursor_page(*, object_type: str):
        controller = __new_controller(object_type)
        request_args = ListGetParamaters(request.args)
        search_after = request.json.get('search_after')
        return controller.get_cursor_page(object_type, request_args, search_after)

    @data_handler.route(
        '/<object_type>:to-one/<object_id>/<path:hops_suffix>',
        methods=['GET']
    )
    def get_to_one_relation(
        *,
        object_type: str,
        object_id: str,
        hops_suffix: str
    ):
        controller = __new_controller(object_type)
        source = data_source_dict[object_type].data_object_factory(
            object_type,
            object_id
        )
        hops = RelataionshipHopsParser(hops_suffix).relationship_hops
        return controller.get_recursive_relation(source, hops)

    @data_handler.route(
        '/<object_type>:to-many/<object_id>/<relationship_name>',
        methods=['GET']
    )
    def get_to_many_relations(
        *,
        object_type: str,
        object_id: str,
        relationship_name: str
    ):
        controller = __new_controller(object_type)
        source = data_source_dict[object_type].data_object_factory(
            object_type,
            object_id
        )
        params = ListGetParamaters(request.args)
        return controller.get_many_relations_page(
            source,
            relationship_name,
            params
        )

    @data_handler.app_errorhandler(BaseRuntimeException)
    def handle_runtime_error(error: BaseRuntimeException):
        return {
            'errors': error.errors
        }, error.status_code

    @data_handler.app_errorhandler(DataSourceError)
    def handle_datasource_error(error: DataSourceError):
        return {
            'errors': [{
                'title': error.title,
                'detail': error.detail
            }]
        }, error.status_code

    @data_handler.app_errorhandler(AuthError)
    def handle_auth_error(error: AuthError):
        return {
            'errors': error.errors
        }, error.status_code

    return data_handler


def data_blueprint(
    *data_sources: DataSource,
    url_prefix: str = '/data',
    config_prefix: str = '/_config',
    auth_inspector: Optional[AuthInspector] = None
) -> DataBlueprint:

    config_bp = _config_blueprint(
        config_prefix,
        data_sources,
        DefaultOperatorConfig(*data_sources)
    )
    core_bp = _core_blueprint(
        DataSourceDict(*data_sources),
        url_prefix,
        auth_inspector=auth_inspector
    )
    core_bp.register_blueprint(config_bp)

    return core_bp


def custom_blueprint(
    url_prefix: str = '/custom',
    name: str = 'custom'
) -> DataBlueprint:
    """
    Provides a flask `Blueprint` instance for adding custom endpoints to.
    """

    custom_handler = CustomBlueprint(name=name, url_prefix=url_prefix)

    return custom_handler



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT


def convert_treeofsex_species_to_data_object(species, target_datasource):
    for s in species:
        CoreDataObject = target_datasource.data_object_factory  # noqa N806
        ret = CoreDataObject(
            'treeofsex_species',
            attributes={
                'higher_taxanomic_group': s['Higher taxonomic group'],
                'order': s['Order'],
                'family': s['Family'],
                'genus': s['Genus'],
                'species': s['Species'],
                'species_author': s['species author'],
                'infraspecific': s['infraspecific'],
                'common_name': s['common name'],
                'hybrid': s['Hybrid'],
                'source_hybrid': s['source: Hybrid'],
                'the_plant_list_v1_1': s['ThePlantListV1.1 (1=accepted at species level,-1=recognized synonym,0=other)'],  # noqa E501
                'name_in_the_plant_list_v1_1': s['Name in ThePlantListV1.1'],
                'sexual_system': s['Sexual system'],
                'source_sexual_system': s['source: Sexual System'],
                'selfing': s['Selfing (self incompatible,self compatible)'],
                'source_selfing': s['source: Selfing'],
                'growth_form': s['Growth Form (herb,shrub,tree,herbaceous vine,liana/woody vine)'],
                'source_growth_form': s['source: Growth Form'],
                'woodiness': s['Woodiness (W, H, variable)'],
                'woodiness_count': s['Woodiness count'],
                'source_woodiness': s['source: Woodiness'],
                'life_form': s['Life Form (annual,perennial)'],
                'source_life_form': s['source: Life Form'],
                'gametophytic_chromosome_num': s['gametophytic chromosome number'],
                'gametophytic_chromosome_num_min': s['gametophytic chromosome number (minimum)'],
                'gametophytic_chromosome_num_mean': s['gametophytic chromosome number (mean)'],
                'source_gametophytic_chromosome_num': s['source: gametophytic chromosome number'],
                'sporophytic_chromosome_num': s['sporophytic chromosome number'],
                'sporophytic_chromosome_num_min': s['sporophytic chromosome number (minimum)'],
                'sporophytic_chromosome_num_mean': s['sporophytic chromosome number (mean)'],
                'source_sporophytic_chromosome_num': s['source: sporophytic chromosome number'],
                'karyotype': s['karyotype (ZO,ZW,XY,XO,WO,homomorphic,complex XY,complex ZW)'],
                'source_karyotype': s['source: karyotype'],
                'molecular_basis': s['molecular basis (dosage,Y dominant,W dominant)'],
                'source_molecular_basis': s['source: molecular basis'],
                'genotypic': s['genotypic (male heterogametic,female heterogametic,GSD,polygenic)'],  # noqa E501
                'source_genotypic': s['source: genotypic'],
                'notes': s['notes,comments'],
                'email': s['entry email'],
                'citation': s['citation']
            }
        )
        yield ret



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import os

from ..api_client import (
    create_api_datasource
)
from ..core import (
    core_data_object
)


def portaldb(retries: int = 5):
    portaldb = create_api_datasource(
        api_url=os.getenv('PORTAL_URL') + os.getenv('PORTAL_API_PATH') + '/local',
        token=os.getenv('PORTAL_API_KEY'),
        data_prefix='',
        retries=retries
    )
    core_data_object(portaldb)
    return portaldb



# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import urllib
from abc import ABC, abstractmethod
from datetime import date, datetime
from typing import Any, Dict, Iterable, List, Optional, Union

from ..core import DataObject
from ..core.operator import Relational

DocumentMeta = Dict[str, Any]
DumpDict = Dict[str, Any]
DumpDictMany = List[DumpDict]
ResponseDict = Dict[str, Union[DumpDict, DumpDictMany]]


class View(ABC):
    """
    Provides an MVC-esque View class. Can serialize both an individual
    DataObject, as well as an Iterable of DataObject instances.
    """

    @abstractmethod
    def dump(
        self,
        data_object: DataObject,
        document_meta: Optional[DocumentMeta] = None
    ) -> ResponseDict:
        """
        Create a JSON:API response for an individual DataObject result
        """

    @abstractmethod
    def dump_bulk(
        self,
        data_objects: Iterable[DataObject],
        document_meta: Optional[DocumentMeta] = None
    ) -> ResponseDict:
        """
        Create a JSON:API response for an Iterable of DataObject results
        """


RelationshipDump = dict[str, dict[str, str]]
AllRelationshipsDump = dict[str, RelationshipDump]


class DefaultView(View):
    """
    Provides a default implementation of the View ABC.
    """

    def __init__(
        self,
        prefix: str = '',
        include_all_to_ones: bool = False,
        hop_limit: Optional[int] = None,
        requested_fields: list[str] | None = None,
    ) -> None:
        """
        Args:

        - prefix                - the URL prefix on which the
                                  data blueprint is served
        - include_all_to_ones   - whether to fetch all absent
                                  to-one relation objects,
                                  using the "host" `Relational`
                                  instance
        - hop_limit             - the maximum recursion limit
                                  on including related to-one
                                  objects. Default no limit
        """

        self.__prefix = prefix
        self.__all_to_ones = include_all_to_ones
        self.__hop_limit = hop_limit
        self.__requested_fields = requested_fields

    def dump(
        self,
        data_object: DataObject,
        document_meta: Optional[DocumentMeta] = None
    ) -> ResponseDict:

        response = {
            'data': self.__dump_object(
                data_object,
                self.__initial_marker,
            )
        }
        if document_meta is not None:
            response['meta'] = document_meta
        return response

    def dump_bulk(
        self,
        data_objects: Iterable[DataObject],
        document_meta: Optional[DocumentMeta] = None
    ) -> ResponseDict:

        dumped = [
            self.__dump_object(
                data_object,
                self.__initial_marker,
            )
            for data_object in data_objects
        ]
        response = {
            'data': dumped
        }
        if document_meta is not None:
            response['meta'] = document_meta
        return response

    @property
    def __initial_marker(self) -> int | str:
        if self.__requested_fields:
            return ''
        else:
            return 0

    def __dump_object(
        self,
        data_object: DataObject,
        marker: int | str,
    ) -> DumpDict:

        dump = {
            'type': data_object.type,
            'id': data_object.id
        }
        if data_object.attributes:
            dump['attributes'] = self.__convert_attributes(
                data_object.attributes
            )
        dump = self.__add_relationships(data_object, dump, marker)
        return dump

    def __add_relationships(
        self,
        data_object: DataObject,
        dump: DumpDict,
        marker: int
    ) -> DumpDict:

        host = data_object._host
        if not isinstance(host, Relational):
            return dump

        to_one_keys = self.__get_to_one_keys(data_object)
        to_many_keys = self.__get_to_many_keys(host, data_object.type)
        if not to_one_keys and not to_many_keys:
            return dump
        dump['relationships'] = self.__get_relationship_dumps(
            to_one_keys,
            to_many_keys,
            data_object,
            marker
        )
        return dump

    def __get_relationship_dumps(
        self,
        to_one_relationships: list[str],
        to_many_relationships: list[str],
        data_object: DataObject,
        marker: int
    ) -> AllRelationshipsDump:

        dump = {
            key: self.__dump_to_one_relationship(key, data_object, marker)
            for key in to_one_relationships
        } | {
            key: self.__dump_to_many_relationship(key, data_object.type,
                                                  data_object.id)
            for key in to_many_relationships
        }

        return {
            k: v for k, v in dump.items() if v != {}
        }

    def __dump_to_many_relationship(
        self,
        key: str,
        type_: str,
        id_: str
    ) -> RelationshipDump:

        id_encoded = urllib.parse.quote(str(id_), safe='')
        link = f'{self.__prefix}/{type_}/{id_encoded}/{key}'
        return {
            'links': {
                'related': link
            }
        }

    def __dump_to_one_relationship(
        self,
        key: str,
        data_object: DataObject,
        marker: int | str
    ) -> Optional[RelationshipDump]:

        if self.__hop_limit is not None and marker >= self.__hop_limit:
            return {}

        if self.__requested_fields and not self.__to_one_is_relevant(key, marker):
            return {}

        related_object = self.__get_related_to_one(data_object, key)
        if related_object is not None:
            next_marker = self.__get_next_marker(marker, key)
            return {
                'data': self.__dump_object(
                    related_object,
                    next_marker,
                )
            }

    def __get_next_marker(
        self,
        marker: int | str,
        key: str,
    ) -> int | str:

        return (
            (f'{marker}.{key}' if marker else key)
            if self.__requested_fields
            else marker + 1
        )

    def __to_one_is_relevant(
        self,
        key: str,
        marker: str
    ) -> bool:

        next_marker = self.__get_next_marker(marker, key)

        return any(
            r for r in self.__requested_fields
            if r.startswith(next_marker)
        )

    def __get_related_to_one(
        self,
        data_object: DataObject,
        key: str
    ) -> Optional[DataObject]:

        relations = (
            data_object.to_one_relationships
            if self.__all_to_ones
            else data_object._to_one_objects
        )
        return relations.get(key)

    def __get_to_one_keys(
        self,
        obj: DataObject
    ) -> list[str]:

        if self.__all_to_ones:
            return self.__get_target_keys(
                obj._host,
                obj.type,
                'to_one'
            )
        else:
            return list(
                obj._to_one_objects.keys()
            )

    def __get_to_many_keys(
        self,
        host: Relational,
        type_: str
    ) -> list[str]:

        return self.__get_target_keys(host, type_, 'to_many')

    def __get_target_keys(
        self,
        host: Relational,
        type_: str,
        target_name: str
    ) -> list[str]:

        if host.relationship_config is None:
            return []
        config = host.relationship_config.get(type_)
        if config is None:
            return []
        target = getattr(config, target_name)
        return self.__keys_or_empty(target)

    def __keys_or_empty(
        self,
        config: Optional[dict[str, str]]
    ) -> Iterable[str]:

        return (
            config.keys() if config is not None else []
        )

    def __convert_attributes(
        self,
        attributes: dict[str, Any]
    ) -> dict[str, Any]:

        return {
            k: self.__convert_value(v)
            for k, v in attributes.items()
        }

    def __convert_value(self, __v: Any) -> Any:
        if isinstance(__v, (date, datetime)):
            return __v.isoformat()
        return __v



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Any, Dict, Optional, Union

from .parser import Parser
from .view import View
from ..core import DataObject
from ..core.relationship import RelationshipConfig


JsonApiObject = dict[str, Any]
JsonApiTransfer = dict[
    str,
    Union[JsonApiObject, list[JsonApiObject]]
]
JsonRelationship = dict[
    str,  # "one" or "many"
    dict[str, str]  # relationship_name:target_type
]
JsonRelationshipConfig = dict[
    str,  # the object_type
    JsonRelationship
]


class JsonApiConverter():

    """
    Converts from JSON:API transfers to instances of
    `DataObject`.
    """

    def __init__(
        self,
        parser: Parser,
        data_key: str = 'data',
        meta_key: str = 'meta'
    ) -> None:

        self.__parser = parser
        self.__data_key = data_key
        self.__meta_key = meta_key

    def convert(self, input_: JsonApiTransfer) -> DataObject:
        """
        Converts a JsonApiTransfer containing a detail (single) result
        """

        json_obj = input_[self.__data_key]
        return self.__parser.parse(json_obj)

    def convert_list(
        self,
        input_: JsonApiTransfer
    ) -> tuple[list[DataObject], Optional[int]]:
        """
        Converts a JsonApiTransfer containing a list of results. Also
        returns a count of the total results meeting.
        """

        json_obj_list = input_[self.__data_key]
        total_count = input_.get('meta', {}).get('total', None)
        return [
            self.__parser.parse(json_obj)
            for json_obj in json_obj_list
        ], total_count

    def convert_count(
        self,
        input_: JsonApiTransfer
    ) -> Dict[str, Any]:
        """
        Converts a JsonApiTransfer containing a list of stats.
        """

        stats = input_[self.__meta_key]
        return stats['total']

    def convert_stats(
        self,
        input_: JsonApiTransfer
    ) -> Dict[str, Any]:
        """
        Converts a JsonApiTransfer containing a list of stats.
        """

        stats = input_[self.__meta_key]
        return self.__parser.parse_stats(stats)

    def convert_group_stats(
        self,
        input_: JsonApiTransfer
    ) -> Dict[str, Any]:
        """
        Converts a JsonApiTransfer containing a list of grouped stats.
        """

        stats = input_[self.__meta_key]
        return self.__parser.parse_group_stats(stats)

    def convert_cursor_page(
        self,
        input_: JsonApiTransfer
    ) -> tuple[list[DataObject], list[str] | None]:
        """
        Converts a `JsonApiTransfer` of a cursor-page
        """

        objs = self.__parser.parse_iterable(
            input_[self.__data_key]
        )
        search_after = input_.get('meta', {}).get('search_after')

        return objs, search_after

    def convert_relationship_config(
        self,
        config_transfer: JsonRelationshipConfig
    ) -> dict[str, RelationshipConfig]:
        """
        Converts a `JsonRelationshipConfig` dict, returned from
        an `api_base2` config blueprint, to a form `ApiDataSource`
        can understand.
        """

        return {
            type_: self.__convert_relationship(rel)
            for type_, rel
            in config_transfer.items()
        }

    def __convert_relationship(
        self,
        rel: JsonRelationship
    ) -> RelationshipConfig:

        return RelationshipConfig(
            to_one=rel.get('one'),
            to_many=rel.get('many')
        )


class DataObjectConverter():

    """
    Converts from instances of `DataObject` to
    JSON:API transfers.
    """

    def __init__(self, view: View) -> None:
        self.__view = view

    def convert(self, input_: DataObject) -> JsonApiTransfer:
        """
        Converts a single `DataObject` instance to a JsonApiTransfer
        """

        return self.__view.dump(input_)

    def convert_list(self, input_: list[DataObject]) -> JsonApiTransfer:
        """
        Converts a `list` of `DataObject` instances to a JsonApiTransfer
        """

        return self.__view.dump_bulk(input_)



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)
from ...core.operator.updater import DataObjectUpdate


class BioscanQcUksiEntryToElasticSampleUpdateConverter(
        DataObjectToDataObjectOrUpdateConverter):

    def convert(self, data_object: DataObject) -> Iterable[DataObjectUpdate]:
        if data_object is not None and data_object.id is not None:

            yield (None, {
                'bold_species': data_object.id,
                'uksi_name_status': data_object.uksi_name_status})



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

def convert_bioscan_barcoding_run_data_to_data_object(brds, target_datasource):
    CoreDataObject = target_datasource.data_object_factory  # noqa N806
    for b in brds:
        ret = CoreDataObject('barcoding_run_data', b)
        yield ret



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import json
from typing import Dict, Optional

from .converter import BoldApiTransfer
from ..core import HttpClient


class BoldApiClient(HttpClient):
    """
    Takes BOLD API transfers and connects to a remote
    BOLD API.
    """

    def __init__(
        self,
        bold_url: str,
        bold_api_key: str,
        retries: int = 5
    ) -> None:
        super().__init__(token=bold_api_key, token_header='api-key', retries=retries)
        self.__bold_url = bold_url

    def get_detail(
        self,
        object_type: str,
        object_ids: str
    ) -> Optional[BoldApiTransfer]:
        """
        Gets a list of BOLD API transfers for the objects of specified
        `object_type` and `object_id`, or returns None if not found.
        """

        url, params = self.__detail_url(object_type, object_ids)
        headers = self._merge_headers()
        return self.__fetch_detail(url, params=params, headers=headers)

    def __fetch_detail(
        self,
        url: str,
        params: Dict = {},
        headers: Dict = {},
    ) -> Optional[BoldApiTransfer]:

        session = self._get_session_with_retries()
        r = session.get(url, params=params, headers=headers)
        if r.status_code in [400, 404]:
            return []
        r.raise_for_status()
        results = []
        lines = r.text.splitlines()
        for line in lines:
            result = json.loads(line)
            results.append(result)
        return results

    def __detail_url(self, object_type: str, object_ids: str) -> str:
        url = f'{self.__bold_url}/records'
        obj_ids_str = ','.join(object_ids)
        params = {
            'sampleids': obj_ids_str
        }
        return url, params



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from abc import ABC, abstractmethod
from collections.abc import Mapping
from typing import Any, Iterable, Optional

from dateutil.parser import parse as dateutil_parse

from ..core import DataObject

if typing.TYPE_CHECKING:
    from ..core import DataSource


JsonApiResource = dict[str, Any]
JsonApiDoc = dict[str, list[JsonApiResource]]


class Parser(ABC):
    """
    Parses JSON:API transfer resource `dict`s to `DataObject`
    instances
    """

    def parse_iterable(
        self,
        transfers: Iterable[JsonApiResource]
    ) -> Iterable[DataObject]:
        """
        Parses an `Iterable` of JSON:API transfer resources
        """

        return (
            self.parse(t) for t in transfers
        )

    @abstractmethod
    def parse(self, transfer: JsonApiResource) -> DataObject:
        """
        Parses an individual JSON:API transfer resource to a
        `DataObject` instance
        """

    @abstractmethod
    def parse_stats(self, transfer: JsonApiResource) -> dict:
        """
        Parses an individual stats transfer resource to a
        stats dict instance
        """

    @abstractmethod
    def parse_group_stats(self, transfer: JsonApiResource) -> list[dict]:
        """
        Parses a grouped stats transfer resource to a
        list instance
        """


class DefaultParser(Parser):

    def __init__(self, data_source_dict: dict[str, DataSource]) -> None:
        self.__dict = data_source_dict

    def parse(self, transfer: JsonApiResource) -> DataObject:
        type_ = transfer['type']
        ds = self.__get_data_source(type_)
        raw_attributes = transfer.get('attributes')

        attributes = self.__convert_attributes(type_, raw_attributes)

        return ds.data_object_factory(
            transfer.get('type'),
            id_=transfer.get('id'),
            attributes=attributes,
            to_one=self.__parse_to_ones(transfer)
        )

    def parse_stats(self, transfer: JsonApiResource) -> dict:
        type_ = transfer.get('type')
        raw_stats = transfer.get('stats')
        converted_stats = self.__convert_stats(type_, raw_stats)
        return {'stats': converted_stats}

    def parse_group_stats(self, transfer: JsonApiResource) -> Iterable[dict]:
        type_ = transfer.get('type')
        raw_stats = transfer.get('stats')

        return [
            self.__convert_group_stats(type_, raw_stat)
            for raw_stat in raw_stats
        ]

    def __get_data_source(self, type_: str) -> DataSource:
        return self.__dict[type_]

    def __parse_to_ones(
        self,
        transfer: JsonApiResource
    ) -> dict[str, DataObject]:

        return {
            k: self.__parse_to_one(v)
            for k, v in transfer.get('relationships', {}).items()
            if self.__relationship_is_to_one(v)
        }

    def __parse_to_one(
        self,
        v: dict[str, Any] | None
    ) -> DataObject | None:

        if v is None:
            return None
        else:
            return self.parse(v.get('data', {}))

    def __relationship_is_to_one(
        self,
        relation: dict[str, Any] | None
    ) -> bool:

        if relation is None:
            return True

        return isinstance(
            relation.get('data'),
            Mapping
        )

    def __convert_attributes(
        self,
        type_: str,
        attributes: Optional[dict[str, Any]]
    ) -> dict[str, Any]:

        if not attributes:
            return {}

        datetime_keys = self.__get_datetime_keys(type_)

        return {
            k: (
                dateutil_parse(v)
                if k in datetime_keys and v is not None
                else v
            )
            for k, v in attributes.items()
        }

    def __convert_stats(
        self,
        type_: str,
        stats: Optional[dict[str, Any]]
    ) -> dict[str, Any]:
        # {'field': {'min': value, 'max': value}
        if not stats:
            return {}

        datetime_keys = self.__get_datetime_keys(type_)

        return {
            fieldname: {
                k: (
                    dateutil_parse(v, ignoretz=True)
                    if fieldname in datetime_keys
                    and v is not None
                    and k in ['min', 'max']
                    else v
                )
                for k, v in fieldstats.items()
            }
            for fieldname, fieldstats in stats.items()
        }

    def __convert_group_stats(
        self,
        type_: str,
        raw_stats: dict[str, dict[str, Any]]
    ) -> dict[str, dict[str, Any]]:

        st = raw_stats.pop('stats')
        count = st.pop('count', None)

        raw_stats['stats'] = self.__convert_stats(type_, st)

        if count is not None:
            raw_stats['stats']['count'] = count

        return raw_stats

    def __get_datetime_keys(self, type_: str) -> list[str]:
        ds = self.__get_data_source(type_)
        attribute_types = ds.attribute_types.get(
            type_,
            {}
        )

        return [
            k for k, v in attribute_types.items()
            if self.__value_is_datetime(v)
        ]

    def __value_is_datetime(self, __v: str) -> bool:
        lower_ = __v.lower()

        return 'date' in lower_ or 'time' in lower_



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT


def convert_to_data_objects(runs, target_datasource):
    CoreDataObject = target_datasource.data_object_factory  # noqa N806
    for run in runs:
        yield CoreDataObject('sequencing_file', run)



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import (
    Iterable
)

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)


class ElasticObjectToPortaldbObjectConverter(
        DataObjectToDataObjectOrUpdateConverter):

    def __init__(self, data_object_factory, fields: dict = {},
                 destination_object_type: str = 'tolid_event',
                 id_field: str = 'id'):
        super().__init__(data_object_factory)
        self.__fields = fields
        self.__destination_object_type = destination_object_type
        self.__id_field = id_field

    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        if data_object is not None:
            ret = self._data_object_factory(
                self.__destination_object_type,
                data_object.get_field_by_name(self.__id_field),
                attributes=self.__fields
            )
            yield ret
        else:
            yield None



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)


class TolqcSampleToElasticSequencingRequestConverter(
        DataObjectToDataObjectOrUpdateConverter):
    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        target_attributes = {**data_object.attributes}

        if data_object.specimen is not None:
            target_attributes['tolid'] = {'id': data_object.specimen.id}
            target_attributes['specimen'] = {'id': data_object.specimen.supplied_name}
            if data_object.specimen.accession is not None:
                target_attributes['biospecimen_id'] = data_object.specimen.accession.id
            if data_object.specimen.species is not None:
                target_attributes['species'] = {'id': data_object.specimen.species.id}

        ret = self._data_object_factory(
            'sequencing_request',
            data_object.id,
            attributes=target_attributes
        )

        return iter([ret])
